# -*- coding: utf-8 -*-
# Author: Andrew.M.G.Reynen
from __future__ import print_function, division
import sys
import logging
import sip
import os
import time
import ctypes
from future.utils import iteritems
from fnmatch import fnmatch
from copy import deepcopy
sip.setapi('QVariant', 2)
sip.setapi('QString', 2)
sys.path.insert(0,os.path.dirname(os.path.realpath(__file__)))
from __init__ import __version__

import numpy as np
from PyQt5 import QtWidgets,QtGui,QtCore
from PyQt5.QtCore import QSettings
from pyqtgraph import mkPen, mkBrush
from obspy import UTCDateTime

from MainWindow import Ui_MainWindow
from CustomWidgets import keyPressToString
from TemporalWidgets import TraceWidget
from CustomFunctions import getTimeFromFileName,getStaStr
from HotVariables import initHotVar
from Preferences import defaultPreferences,DateDialog
from Actions import defaultActions,defaultPassiveOrder,QueueThread
from Archive import getArchiveAvail,extractDataFromArchive
from StationMeta import staXml2Loc,readInventory,projStaLoc
from ConfigurationDialog import ConfDialog
from SaveSource import CsDialog,defaultSource

# Main window class
class LazylystMain(QtWidgets.QMainWindow, Ui_MainWindow):
    def __init__(self):
        QtWidgets.QMainWindow.__init__(self)
        Ui_MainWindow.__init__(self)
        self.setupUi(self)
        self.tweakUi()
        self.loadSettings()
        self.applyPreferences()
        self.setFunctionality()
        self.introduction()
        self.processAction(self.act['OpenLazylyst'])
    
    # Update UI for the few commands not captured in Qt Designer
    def tweakUi(self):
        self.traceLayout.setContentsMargins(0,0,0,0)
        self.timeImageLayout.setContentsMargins(0,0,0,0)
        self.setWindowTitle('Lazylyst '+__version__)
     
    # Go through all preferences, and call their update functions
    def applyPreferences(self):
        curPrefs=[aPref for key, aPref in iteritems(self.pref)]
        for aPref in curPrefs:
            aPref.update(self,init=True)
        # Also check initial image widget visibility, has to be called after staWidgets are added
        if self.setGen.value('imageHidden','false')=='true':
            self.toggleImageWidget(init=True)
        
    # Start setting up some functionality to the UI
    def setFunctionality(self):
        # Connect the zoomed in and zoomed out views of the event files
        self.archiveSpan.span.sigRegionChanged.connect(lambda: self.archiveEvent.updateXRange(self.archiveSpan.span))
        self.archiveSpan.span.sigRegionChanged.connect(self.updateSpanText)
        self.archiveSpan.span.sigRegionChangeFinished.connect(self.updateArchiveSpanList)
        # Allow user to set manually the ranges (no need to use slider)
        self.archiveSpanT0Label.doubleClicked.connect(lambda: self.setSpanBoundViaDialog('T0'))
        self.archiveSpanT1Label.doubleClicked.connect(lambda: self.setSpanBoundViaDialog('T1'))
        # Give ability to the archive displays
        self.archiveEvent.addNewEventSignal.connect(lambda: self.addPickFile(self.archiveEvent.newEveTime))
        self.archiveSpan.addNewEventSignal.connect(lambda: self.addPickFile(self.archiveSpan.newEveTime))
        self.archiveList.doubleClicked.connect(self.archiveListDoubleClickEvent)
        self.archiveListLineEdit.editingFinished.connect(self.updateArchiveSpanList)
        # Give ability to the map
        self.mapWidget.doubleClicked.connect(self.mapDoubleClickEvent)
        self.mapWidget.updatePolygonPenSignal.connect(self.updateMapPolygonPen)
        # Link the image axis to the time axis
        self.imageWidget.setXLink('timeAxis')
        # Allow stdout to be sent to the Trace Log
        XStream.stdout().setTextCursor.connect(self.setTextCursor)
        XStream.stderr().setTextCursor.connect(self.setTextCursor)
        XStream.stdout().messageWritten.connect(self.textOutBrowser.insertPlainText)
        XStream.stderr().messageWritten.connect(self.textOutBrowser.insertPlainText)
        self.textOutBrowser.textChanged.connect(self.scrollTextOut)
        
    # Move the text browser cursor to the bottom
    def setTextCursor(self):
        self.textOutBrowser.moveCursor(QtGui.QTextCursor.End)
    
    # Scroll the trace log to the bottom
    def scrollTextOut(self):
        # Check first to see if the text should be trimmed
        splitText=self.textOutBrowser.toPlainText().split('\n')
        lineCount=len(splitText)
        maxCount,lineBuff=250,50
        if lineCount>maxCount+lineBuff:
            # Disconnect from itself - no need to scroll twice, then update the trimmed text
            self.textOutBrowser.textChanged.disconnect(self.scrollTextOut)
            self.textOutBrowser.setPlainText('\n'.join(splitText[-maxCount:]))
            self.textOutBrowser.textChanged.connect(self.scrollTextOut)
        # Scroll to bottom of log
        scrollBar=self.textOutBrowser.verticalScrollBar()
        scrollBar.setValue(scrollBar.maximum())

    # Toggle the image widget on or off
    def toggleImageWidget(self,init=False):
        dock=self.imageWidget
        maxHeight=self.timeWidget.maximumHeight()
        if dock.isHidden():
            dock.show()
            if self.traceSplitSizes is not None:
                self.traceSplitter.setSizes(self.traceSplitSizes)
            else:
                self.traceSplitter.setSizes([maxHeight+70,np.sum(self.traceSplitter.sizes())-(maxHeight+70)])
        else:
            # If starting up lazylyst, force the splitter boundary
            if init:
                self.traceSplitter.setSizes([maxHeight,999999])
            else:
                self.traceSplitSizes=self.traceSplitter.sizes()
                self.traceSplitter.setSizes([maxHeight,np.sum(self.traceSplitSizes)-maxHeight])
            dock.hide()
        ## Do not allow resizing splitter if there is no image being shown ##
            
    # Reload plugins from the specified modules
    def reloadPlugins(self):
        for key,action in iteritems(self.act):
            if action.path=='$main':
                continue
            action.linkToFunction(self,reloadMod=True)
        print('Reloaded plugins')
    
    # Report the current keybinds for the configuration and change source actions
    def introduction(self):
        print('Press: '+self.act['ChangeSource'].trigger.toString()+' to change source (archive/picks/station metadata)')
        print('Press: '+self.act['OpenOptions'].trigger.toString()+' to open the configuration dialog')
    
    # Function to handle key board input from user
    def keyPressEvent(self, ev):
        super(LazylystMain, self).keyPressEvent(ev)
        keyname=keyPressToString(ev)
        if keyname==None:
            return
        # Set the time the user pressed this button
        self.setUserSeenAtTime()
        # Loop through all actions and see if one is activated...
        # ... use the original set (an action may be added part way through, so can mess with the loop)
        actions=[action for key,action in iteritems(self.act) if action.trigger!='DoubleClick']
        for action in actions:
            if not action.passive and action.trigger.toString()==keyname:
                self.processAction(action)

    # Function to handle double click events from trace widgets
    def traceDoubleClickEvent(self):
        action=self.act['PickAdd']
        self.processAction(action)
    
    # Function to handle double clicks from the archive widget
    def archiveListDoubleClickEvent(self):
        action=self.act['PickFileSetToClick']
        self.processAction(action)
    
    # Function to handle double clicks from the map widget
    def mapDoubleClickEvent(self):
        action=self.act['MapDblClicked']
        self.processAction(action)
    
    # Create and activate the queue of actions following the initiation of an active action
    def processAction(self,action):
        # If the action is sleeping, do nothing
        if action.sleeping:
            print('Action '+action.tag+' is sleeping')
            return
        # If this action is timed...
        if action.timer:
            # ... see if already being run, and stop if so
            curTimers=self.qTimers.keys()
            if action.tag in curTimers:
                self.stopTimedAction(action)
            # ... otherwise add to the timers, and kick it off
            else:
                timer = QtCore.QTimer()
                timer.timeout.connect(lambda: self.runActiveAction(action))
                self.qTimers[action.tag]=timer
                self.qTimers[action.tag].start(action.timerInterval*1000.0)
                self.updateStrollingList()
        # If not timed, just do it once
        else:
            self.runActiveAction(action)

    # Run the specified action
    def runActiveAction(self,action):
        # Set the current station, current trace ranges, and afk time to be sent to actions
        self.setCurTraceStaAndPos()
        self.setTraceRanges()
        self.setAfkTime()
        # Check to see if there are any (passive) actions which relate
        actQueue=self.collectActQueue(action)
        # For potentially lengthy updates, only do so if necessary
        seenTags=set([tag for act in actQueue for tag in self.act[act.tag].inputs])
        for tag,func in [['mapPolygon',self.setMapPolygon]]:
            if tag in seenTags:
                func()
        # If the trigerring action is threaded, send the queue to a thread 
        if action.threaded: 
            # First check to see if the thread is already running, skip if it is
            if action.tag in self.qThreads.keys():
                print('A thread is already running which was initiated by '+action.tag)
            else:
                thread=QueueThread(action.tag,actQueue)
                # Mark this thread, given the active actions tag
                self.qThreads[action.tag]=thread
                self.updateSchemingList()
                # Connect to the threads signals for when to send inputs, and collect returns
                thread.setNextInputs.connect(lambda: self.setThreadInputs(thread))
                thread.sendReturns.connect(lambda: self.updateReturns(thread.actQueue[thread.curIdx],thread.returns,thread.tag))
                thread.finished.connect(lambda: self.updateThreadDict(thread.tag))
                thread.start()
        else:
            for oAct in actQueue:
                # Collect the required inputs
                inputs=self.collectActInputs(oAct.tag)
                # Call the function with args and kwargs
                returnVals=oAct.func(*inputs,**oAct.optionals)
                self.updateReturns(oAct,returnVals,action.tag)
    
    # Update the threads inputs to be its next to process actions inputs
    def setThreadInputs(self,thread):
        thread.setInputs(self.collectActInputs(thread.actQueue[thread.curIdx].tag))
    
    # When the thread has finished, remove the triggers tag from the thread dictionary
    def updateThreadDict(self,threadTag):
        self.qThreads.pop(threadTag)
        self.updateSchemingList()
                
    # Get the appropriate order of passive functions before and after their triggered active action
    def collectActQueue(self,action):
        beforeActive=[] # Passive actions with beforeTrigger
        afterActive=[] # Passive actions without beforeTrigger
        # Collect all the passive actions which have this active action as a trigger...
        # ...in the order specified within the configuration list
        for actTag in self.actPassiveOrder:
            if action.tag not in self.act[actTag].trigger:
                continue
            # If the actions function wasn't initialized, or is sleeping, skip
            if self.act[actTag].func==None or self.act[actTag].sleeping:
                continue
            if self.act[actTag].beforeTrigger:
                beforeActive.append(self.act[actTag])
            else:
                afterActive.append(self.act[actTag])
        return beforeActive+[action]+afterActive
    
    # Collect all of the inputs to be sent to a specific action
    def collectActInputs(self,actionTag):
        # Collect the required inputs
        inputs=[]
        hotVarKeys=self.hotVar.keys()
        for key in self.act[actionTag].inputs:
            if key in hotVarKeys:
                inputs.append(self.hotVar[key].getVal())
            else:
                inputs.append(self.pref[key].getVal())
        return inputs
        
    # Update all return (hot variable) values from an action which just finished executing
    def updateReturns(self,action,returnVals,triggerTag):
        # If no returns, but got something, let user know...
        if len(action.returns)==0:
            if returnVals is not None:
                print('Action '+action.tag+' expected no returns, but received some')
            return
        # ...if just one return, convert to list to treat it the same as multiple returns
        elif len(action.returns)==1:
            returnVals=[returnVals]
        # ... if expecting multiple returns, ensure the length of the return value array can be taken
        else:
            try:
                len(returnVals)
            except:
                returnVals=[returnVals]
        # Go through each of the returns in order, and update...
        # ...check first to see the number of returns are correct
        # ...and that the previous/new types match
        # ...pass (ignore) updates on variables with "$pass" as the return value
        if len(action.returns)==len(returnVals):
            skipUpdates=False
            # Ensure all the variable types match before updating any hot variables...
            # ...also ensure that the hot variables ("sanity") check function passes
            for i,aReturnKey in enumerate(action.returns):
                # Do not do any checks if the user wanted to pass this returns value
                if str(returnVals[i])=='$pass':
                    continue
                if not self.testReturnType(self.hotVar[aReturnKey].dataType,type(returnVals[i])):
                    print('Action '+action.tag+' expected variable '+str(self.hotVar[aReturnKey].dataType)+
                           ' for hot variable '+aReturnKey+', got '+str(type(returnVals[i])))
                    skipUpdates=True
                    # If the type is wrong, no point in the longer check
                    continue
                # Test if the return variable passes the built in check
                if self.hotVar[aReturnKey].check==None:
                    continue
                elif not self.hotVar[aReturnKey].check(self,returnVals[i]):
                    print('Action '+action.tag+' failed '+aReturnKey+' check')
                    skipUpdates=True
            if skipUpdates: 
                # Stop the action if it is timed
                if action.timer:
                    self.stopTimedAction(action)
                return
            # Process the return keys in order ...
            for i,aReturnKey in enumerate(action.returns):
                # ... passing over any which have $pass as the return value
                if str(returnVals[i])=='$pass':
                    pass
                else:
                    self.hotVar[aReturnKey].val=returnVals[i]
                    self.hotVar[aReturnKey].update()
        else:
            print('For action '+action.tag+' got '+str(len(returnVals))+
                   ' return values, expected '+str(len(action.returns)))
        # If this was an action which was threaded, reset the return value of the thread...
        # ...so that the thread knows when to continue processing its queues other actions
        if triggerTag in self.qThreads.keys():
            self.qThreads[triggerTag].resetInputsAndReturns()
    
    # As some variable types are quite similar, allow some to be treated the same...
    # ... for now just str and np.string_
    def testReturnType(self,wantType,returnType):
        # Accepted variations for strings
        if wantType==str and returnType in [str,np.string_]:
            return True
        # All other data types
        elif wantType==returnType:
            return True
        return False        
            
    # Remove a timed action from the queue
    def stopTimedAction(self,action):
        self.qTimers[action.tag].stop()
        self.qTimers.pop(action.tag)
        self.updateStrollingList()
    
    # Set the user seen at time to the current time...
    # ...called when user presses a button
    def setUserSeenAtTime(self):
        self.userSeenAtTime=time.time()
    
    # Open the configuration window
    def openConfiguration(self):
        self.dialog=ConfDialog(actions=self.act,main=self,
                               pref=self.pref,hotVar=self.hotVar)
        self.dialog.exec_()
        
    # Open the change source window
    def openChangeSource(self):
        self.dialog=CsDialog(self.hotVar,self.saveSource)
        # Extract the wanted source and activate (does not require to be a saved source)
        if self.dialog.exec_():
            source=self.dialog.returnSource()
            self.hotVar['sourceTag'].val=source.tag
            self.setLazySource(source)
    
    # Update the source to a specific source tag, given via hot variable update
    def updateSource(self):
        source=self.saveSource[self.hotVar['sourceTag'].val]
        self.setLazySource(source)

    # Set the specified source
    def setLazySource(self,source):
        # If the files exist, update the hot variables
        if source.pathExist():
            for key,val in [['archDir',source.archDir],
                            ['pickDir',source.pickDir],
                            ['staFile',source.staFile],
                            ['sourceDict',source.sourceDict]]:
                self.hotVar[key].val=val
                self.hotVar[key].update()
        else:
            print('Source update skipped')

    # Add an empty pick file to the pick directory, also add to GUI list
    def addPickFile(self,pickFileTime):
        # Ignore if no pick directory is currently set
        if self.hotVar['pickDir'].val=='':
            return
        try:
            aTimeStr=UTCDateTime(pickFileTime).strftime('%Y%m%d.%H%M%S.%f')
        except:
            print('Expected new pick file time in format timestamp(s)')
            return
        # Assign an ID to the new event
        seenIDs=[int(aFile.split('_')[0]) for aFile in self.hotVar['pickFiles'].val]
        # Use the specified ID generation style
        if self.pref['eveIdGenStyle'].val=='fill':
            newID=0
            while newID in seenIDs:
                newID+=1
        elif self.pref['eveIdGenStyle'].val=='next':
            if len(seenIDs)==0:
                newID=0
            else:
                newID=np.max(seenIDs)+1
        else:
            print('The eveIDGenStyle '+self.pref['eveIdGenStyle'].val+' has not been implemented')
            return
        # Use the previous file ID length, default at 10 if no files present
        zFillLen=10
        if len(seenIDs)>0:
            zFillLen=len(self.hotVar['pickFiles'].val[0].split('_')[0])
        newPickFile=str(newID).zfill(zFillLen)+'_'+aTimeStr+'.picks'
        # Add to the picks directory
        newFile=open(self.hotVar['pickDir'].val+'/'+newPickFile,'w')
        newFile.close()
        # Add to GUI list, and internal list
        self.hotVar['pickFiles'].val=self.getPickFiles()
        self.updateEveSort()
        self.archiveEvent.updateEveLines(self.hotVar['pickFileTimes'].val,self.hotVar['curPickFile'].val)
        self.updateArchivePrevEvePen()
        self.updateArchiveCurEvePen()
    
    # Load a specific pick file from the pick directory (inter-event pick loading)...
    # ...hot variable pickSet is reset just prior to calling this (with no picks)
    def loadPickSet(self):
        path=self.hotVar['pickDir'].val+'/'+self.hotVar['curPickFile'].val
        # Check that the path exists
        if not os.path.exists(path):
            print('Pick file at '+path+' no longer exists')
            return
        # Check that the file has some content
        elif os.path.getsize(path)==0:
            return # pickSet is set to empty prior
        pickSet=np.genfromtxt(path,delimiter=',',dtype=str)
        # In the odd case there are empty spaces in the file
        if len(pickSet)==0:
            return
        # Put into proper dimensions if only one pick was present
        if len(pickSet.shape)==1:
            pickSet=pickSet.reshape((1,3))
        # Check to see that this file was not tampered with
        if not self.hotVar['pickSet'].check(self,pickSet):
            print('Pick file at '+path+' was not in correct format')
            return
        # Update the pickset
        self.hotVar['pickSet'].val=pickSet
        
    # Save the current picks
    def savePickSet(self):
        # If the current pick file was not yet initiated, nothing to save
        if self.hotVar['curPickFile'].val=='':
            return
        np.savetxt(self.hotVar['pickDir'].val+'/'+self.hotVar['curPickFile'].val,
                   self.hotVar['pickSet'].val,fmt='%s',delimiter=',')
        
    # Set the current pick file, called from double click an event in the archive list
    def setCurPickFileOnClick(self):
        return str(self.archiveList.currentItem().text())
        
    # Load the event related information
    def updateEvent(self):
        # Clear away the previous set of picks
        self.hotVar['pickSet'].val=np.empty((0,3))
        # Go back to the first page
        self.hotVar['curPage'].val=0
        # Load the picks from given pick file, if it has been initialized...
        if self.hotVar['curPickFile'].val!='':
            self.loadPickSet()
            # Ensure that the archive span and list includes this file
            self.checkArchiveSpan()
            # Get the wanted time, and query for the data
            aTime=getTimeFromFileName(self.hotVar['curPickFile'].val).timestamp
            t1,t2=aTime+self.pref['evePreTime'].val,aTime+self.pref['evePostTime'].val
            self.hotVar['stream'].val=extractDataFromArchive(t1,t2,self.hotVar['archFiles'].val,
                                                             self.hotVar['archFileTimes'].val)
            # Get the trace background coloring
            self.traceBgColors=self.getStaColors('traceBg')
            # Make a copy for any filtering to be applied
            self.hotVar['pltSt'].val=self.hotVar['stream'].val.copy()
            # Sort traces by channel so they are added in same order (relative other stations)
            self.hotVar['pltSt'].val.sort(keys=['channel'])
            # Alphabetical sorting by default
            self.hotVar['staSort'].val=np.sort(np.unique([getStaStr(tr) for tr in self.hotVar['stream'].val]))
            # If there is data...
            # ...move the axis time limit to the appropriate position
            if len(self.hotVar['stream'].val)!=0:
                minTime=np.min([tr.stats.starttime.timestamp for tr in self.hotVar['stream'].val])
                maxTime=np.max([tr.stats.endtime.timestamp for tr in self.hotVar['stream'].val])
                self.timeWidget.plotItem.setXRange(minTime,maxTime)
            # ...let user know otherwise
            else:
                print('No data extracted from archive for curPickFile')
        # ...Otherwise reset hot variable values relating to curPickFile
        else:
            defaultHot=initHotVar()
            for key in ['stream','pltSt','staSort','pickSet','curTraceSta','curTracePos']:
                self.hotVar[key].val=defaultHot[key].val
        # Add data, and picks to the station widgets
        self.updatePage()
        # Update the highlighted event on the archive visual
        self.archiveEvent.updateEveLineSelect(self.hotVar['curPickFile'].val)
        self.updateArchiveCurEvePen()
    
    # Update the data and picks on the current page
    def updatePage(self,init=False):
        # Update the title on the time widget
        title=self.hotVar['curPickFile'].val
        if title!='' and len(self.hotVar['staSort'].val)!=0:
            title+=(' Page '+str(self.hotVar['curPage'].val+1)+' of '+
                    str(int(np.ceil(len(self.hotVar['staSort'].val)/float(self.pref['staPerPage'].val)))))
        self.timeWidget.getPlotItem().setLabels(title=title)
        # Update the trace curves
        self.updateTraces()
        # Update the picks
        self.updatePagePicks()
    
    # Function to handle updates of the hot variable curPage
    def updateCurPage(self):
        maxPageNum=(len(self.hotVar['staSort'].val)-1)//self.pref['staPerPage'].val
        if maxPageNum==-1:
            return
        # Clip to the max/minimum if outside the accepted range
        if self.hotVar['curPage'].val>maxPageNum:
            self.hotVar['curPage'].val=maxPageNum
        elif self.hotVar['curPage'].val<0:
            self.hotVar['curPage'].val=0
        self.updatePage()
    
    # Update the time range upon user request
    def updateTimeRange(self):
        t0,t1=self.hotVar['timeRange'].val
        self.timeWidget.setXRange(t0,t1,padding=0.0)
        
    # Update the y-range for all staWidgets upon user request
    def updateTraceRangesY(self):
        yRanges=self.hotVar['yTraceRanges'].val
        for i,widget in enumerate(self.staWidgets):
            widget.setYRange(yRanges[i,0],yRanges[i,1],padding=0.0)
        
    # Built in function to tab to the next or previous page number
    def tabCurPage(self,nextPage=False,prevPage=False,pageNum=0):
        maxPageNum=(len(self.hotVar['staSort'].val)-1)//self.pref['staPerPage'].val
        curNum=self.hotVar['curPage'].val
        # If the next or previous page, ensure still in bounds
        if nextPage and self.hotVar['curPage'].val+1<=maxPageNum:
            self.hotVar['curPage'].val+=1
        elif prevPage and self.hotVar['curPage'].val-1>=0:
            self.hotVar['curPage'].val-=1
        else:
            return
        # If the page number didnt change, do nothing
        if curNum==self.hotVar['curPage'].val:
            return
        # If got to the end, the page number must have changed, update the page
        self.updatePage()
    
    # If the current pick file moved outside of the selected range...
    # ...increase the range to include the pick file
    def checkArchiveSpan(self):
        fileTime=getTimeFromFileName(self.hotVar['curPickFile'].val).timestamp
        t1,t2=self.archiveSpan.span.getRegion()
        reset=False
        if fileTime<t1:
            t1=fileTime-0.02*(t2-t1)
            reset=True
        if fileTime>t2:
            t2=fileTime+0.02*(t2-t1)
            reset=True
        # Move the span select, and the archive time limits
        if reset:
            self.archiveSpan.span.setRegion((t1,t2))
            self.archiveSpan.setXRange(t1,t2) # Padding is already included here
        
    # Set which station is currently being hovered over
    def setCurTraceStaAndPos(self):
        for widget in self.staWidgets:
            if widget.hasFocus():
                self.hotVar['curTraceSta'].val=widget.sta
                self.hotVar['curTracePos'].val=widget.hoverPos
                return
        self.hotVar['curTraceSta'].val=''
    
    # Set the timeRange variable to the current time range of the time widget...
    # ...as well as all of the current y-limits on the staWidgets
    def setTraceRanges(self):
        self.hotVar['timeRange'].val=self.timeWidget.getTimeRange()
        self.hotVar['yTraceRanges'].val=np.array([widget.getRangeY() for widget in self.staWidgets],dtype=float)
    
    # Set the time since the last user interaction
    def setAfkTime(self):
        self.hotVar['afkTime'].val=time.time()-self.userSeenAtTime
    
    # Set the mapPolygon variable to have the currently displayed polygon vertices        
    def setMapPolygon(self):
        self.hotVar['mapPolygon'].val=self.mapWidget.getPolygonVertices()
        
    # Add a pick to the double-clicked station (single-pick addition)
    def addClickPick(self):
        if self.hotVar['curPickFile'].val=='':
            return
        # Return if no pick mode selected
        if self.hotVar['pickMode'].val=='':
            print('No pick mode selected')
            return
        # Return if the selected pick mode is not in the accepted types
        if self.hotVar['pickMode'].val not in self.pref['pickTypesMaxCountPerSta'].val.keys():
            print('Pick mode '+self.hotVar['pickMode'].val+' is not defined in preference pickTypesMaxCountPerSta')
            return
        # Figure out which widget was picked on
        aList=[aWidget for aWidget in self.staWidgets if aWidget.hasFocus()]
        if len(aList)!=1:
            print('No trace widget is in focus, skipped')
            return
        widget=aList[0]
        curMode=self.hotVar['pickMode'].val
        # Append this pick to the pick set, and plotted lines
        aPosX=widget.hoverPos[0]
        # If the position was not set, skip
        if aPosX==None:
            print('Trace widget in focus, but position not set (bug!)')
            return
        self.hotVar['pickSet'].val=np.vstack((self.hotVar['pickSet'].val,
                                              [widget.sta,curMode,aPosX]))
        widget.addPick(aPosX,curMode,self.getPickPen(curMode))
        # Remove picks from the plot and pickSet, where there are too many
        self.remExcessPicks(checkStas=[widget.sta],checkTypes=[curMode])

    # Remove excess picks, which occurs when a given pick type has more than "X"...
    # ...number of picks on a given station (see pref['pickTypesMaxCountPerSta'])
    def remExcessPicks(self,checkStas=None,checkTypes=None):
        # If sta or types is present, check only to delete those pickTypes on given stations
        if checkStas==None:
            checkStas=self.hotVar['staSort'].val
        if checkTypes==None:
            checkTypes=self.pref['pickTypesMaxCountPerSta'].val.keys()
        delIdxs=[]
        visStas=[widget.sta for widget in self.staWidgets]
        # Gather all indicies to be removed from pickSet and pltLines
        for pickType in checkTypes:
            for sta in checkStas:
                potIdxs=np.where((self.hotVar['pickSet'].val[:,0]==sta)&
                                 (self.hotVar['pickSet'].val[:,1]==pickType))[0]
                pickCountMax=self.pref['pickTypesMaxCountPerSta'].val[pickType]
                # If there are more picks than allowed, add to delete indices
                if len(potIdxs)>pickCountMax:
                    # Oldest: picks at the top of the pickSet are removed first
                    if self.pref['remExcessPicksStyle'].val=='oldest':
                        delIdxs+=list(np.sort(potIdxs)[:-pickCountMax])
                    # Closest/furthest: picks are removed relative the current hovered position
                    else:
                        # Not considering removing the most recent pick
                        numToDel=len(potIdxs)-pickCountMax
                        potIdxs=potIdxs[:-1]
                        sortDistArgs=np.argsort(np.abs(self.hotVar['pickSet'].val[potIdxs,2].astype(np.float64)
                                      -np.float64(self.hotVar['curTracePos'].val[0])))
                        if self.pref['remExcessPicksStyle'].val=='closest':
                            delIdxs=potIdxs[sortDistArgs[:numToDel]]
                        else:
                            delIdxs=potIdxs[sortDistArgs[::-1][:numToDel]]
                    delTimes=self.hotVar['pickSet'].val[delIdxs,2]
                    # Delete the excess lines from the widget (if currently shown)
                    if sta in visStas:
                        self.staWidgets[visStas.index(sta)].removePicks(pickType,delTimes)
        # Remove the picks from the hot variable pickSet
        keepIdxs=np.array([i for i in np.arange(len(self.hotVar['pickSet'].val)) if i not in delIdxs])
        self.hotVar['pickSet'].val=self.hotVar['pickSet'].val[keepIdxs]
        
    # Function to handle updates to the hot variable pickSet (assumes all picks have changed)
    # ... adding and remove picks on the current page (does not remove excess)
    def updatePagePicks(self,init=False):
        # If starting up, no picks are present so skip
        if init:
            return
        # Find out which of these values are unique (to remove duplicate picks)
        curList=[str(a[0])+str(a[1])+str(a[2]) for a in self.hotVar['pickSet'].val]
        unqCur,unqCurIdx=np.unique(curList,return_index=True)
        # Update the new set with only the unique picks
        self.hotVar['pickSet'].val=self.hotVar['pickSet'].val[unqCurIdx]
        # Remove any old plot lines
        for widget in self.staWidgets:
            for line in widget.pickLines:
                widget.pltItem.removeItem(line)
            widget.pickLines=[]
        # Add in the new lines
        stas=[widget.sta for widget in self.staWidgets]
        for sta,aType,aTime in self.hotVar['pickSet'].val:
            if sta not in stas:
                continue
            self.staWidgets[stas.index(sta)].addPick(aTime,aType,self.getPickPen(aType))
    
    # Return the wanted pick types pen, in RGB
    def getPickPen(self,aType):
        # If this phases color is not present, apply the default
        if aType in self.pref['pickPen'].val.keys():
            col,width,depth=self.pref['pickPen'].val[aType]
        else:
            col,width,depth=self.pref['pickPen'].val['default']
        col=QtGui.QColor(col)
        return (col.red(), col.green(), col.blue()),width,depth
    
    # Match all streams channels to its appropriate pen
    def getStreamPens(self):
        penRef={} # Dictionary to return with all unique channels pens
        acceptKeys=[key for key in self.pref['customPen'].val.keys()]
        unqChas=np.unique([tr.stats.channel for tr in self.hotVar['stream'].val])
        # First check what tracePenAssign keys are actually accepted
        useKeys=[]
        for key in self.hotVar['tracePenAssign'].val.keys():
            # If the user returned default, let them know it is used if others do not exist
            if key=='default':
                print('customPen tag default is used as a catch all, ignored so does not overpower other trace pens')
            elif key in acceptKeys:
                useKeys.append(key)
            else:
                print('customPen tag '+key+ ' is not currently defined in preference customPen, applying default')
        # Loop through each unique channel and add its pen to penRef
        for cha in unqChas:
            colorInt,width=None,None
            matched=False
            # Check each of the entries for this key for a match
            for key,aList in [[aKey,self.hotVar['tracePenAssign'].val[aKey]] for aKey in useKeys]:
                for entry in aList:
                    if fnmatch(cha,entry):
                        colorInt,width,depth=self.pref['customPen'].val[key]
                        matched=True
                        break
                if matched:
                    break
            # If there was no match, apply default
            if colorInt==None:
                colorInt,width,depth=self.pref['customPen'].val['default']
            pen=mkPen(QtGui.QColor(colorInt),width=width)
            penRef[cha]=[pen,depth]
        return penRef
    
    # Gather the color assignment on a per station basis
    def getStaColors(self,assignType):
        # Only set the pen reference for stations which are present
        if assignType=='mapSta':
            if self.hotVar['staLoc'].val.shape[0]==0:
                unqStas=[]
            else:
                unqStas=np.unique(self.hotVar['staLoc'].val[:,0])
            defaultColInt=self.pref['basePen'].val['mapStaDefault'][0]
            assignKey='mapStaPenAssign'
        else:
            unqStas=np.unique([getStaStr(tr) for tr in self.hotVar['stream'].val])
            defaultColInt=self.pref['basePen'].val['traceBackground'][0]
            assignKey='traceBgPenAssign'
        # Give empty staWidgets the default color
        colorRef={'':QtGui.QColor(defaultColInt)}
        # See which colors have been set
        useKeys=[]
        for key in self.hotVar[assignKey].val.keys():
            if key not in self.pref['customPen'].val.keys():
                print('customPen tag '+key+ 'is not currently defined in preference customPen, applying default')
            else:
                useKeys.append(key)
        # Assign stations colors (if none present, use defaultColInt)
        # Loop through each station and add its color to colorRef
        for sta in unqStas:
            colorInt,matched=None,False
            # Check each of the entries for this key for a match
            for key,aList in [[aKey,self.hotVar[assignKey].val[aKey]] for aKey in useKeys]:
                for entry in aList:
                    if fnmatch(sta,entry):
                        colorInt=self.pref['customPen'].val[key][0]
                        matched=True
                        break
                if matched:
                    break
            # If there was no match, apply default
            if colorInt==None:
                colorInt=defaultColInt
            colorRef[sta]=QtGui.QColor(colorInt)
        return colorRef
    
    # Update the trace curves on the current page
    def updateTraces(self):
        # As new channels could have been added, update the pen reference
        self.chaPenRef=self.getStreamPens()
        # Clear away all previous curves, and references to this station
        for i in np.arange(len(self.staWidgets)):
            for curve in self.staWidgets[i].traceCurves:
                self.staWidgets[i].removeItem(curve)
            self.staWidgets[i].traceCurves=[]
            self.staWidgets[i].pltItem.setLabel(axis='left',text='empty')
            self.staWidgets[i].sta=''
        # Add in the trace data for the current page
        i=0
        stas=np.array([getStaStr(tr) for tr in self.hotVar['pltSt'].val])
        numStas=len(np.unique(stas))
        while self.hotVar['curPage'].val*self.pref['staPerPage'].val+i<numStas:
            if i==self.pref['staPerPage'].val:
                break
            # Figure out which traces are associated with the next station in staSort
            thisSta=self.hotVar['staSort'].val[self.hotVar['curPage'].val*self.pref['staPerPage'].val+i]
            wantIdxs=np.where(stas==thisSta)[0]
            # Also set the y-limits
            ymin=np.min([np.min(self.hotVar['pltSt'].val[idx].data) for idx in wantIdxs])
            ymax=np.max([np.max(self.hotVar['pltSt'].val[idx].data) for idx in wantIdxs])
            self.staWidgets[i].setYRange(ymin,ymax)
            # Assign the widget a station code
            self.staWidgets[i].sta=thisSta
            self.staWidgets[i].pltItem.setLabel(axis='left',text=thisSta)
            # Plot the data
            for idx in wantIdxs:
                trace=self.hotVar['pltSt'].val[idx]
                pen,depth=self.chaPenRef[trace.stats.channel]
                self.staWidgets[i].addTrace(trace.times()+trace.stats.starttime.timestamp,trace.data,
                                            trace.stats.channel,pen,depth)
            i+=1
        # Set the background color
        for widget in self.staWidgets:                        
            widget.setBackground(self.traceBgColors[widget.sta])
    
    # Update all custom pens
    def updateCustomPen(self,init=False):
        self.updateTracePen()
        self.updateTraceBackground()
        self.updateMapStations()
    
    # Update just the trace pens on the current page
    def updateTracePen(self):
        self.chaPenRef=self.getStreamPens()
        for widget in self.staWidgets:
            for curve in widget.traceCurves:
                pen,depth=self.chaPenRef[curve.cha]
                curve.setPen(pen)
                curve.setZValue(depth)
                # If width of zero, do not show the curve
                if pen.widthF()==0:
                    curve.setVisible(False)
                else:
                    curve.setVisible(True)
    
    # Update the default widget colors
    def updateBaseColors(self,init=False):
        for key,func in [['widgetText',self.updateTextColor],
                         ['traceBackground',self.updateTraceBackground],
                         ['timeBackground',self.updateTimeBackground],
                         ['imageBackground',self.updateImageBackground],
                         ['imageBorder',self.updateImageBorderPen],
                         ['mapBackground',self.updateMapBackground],
                         ['mapStaDefault',self.updateMapStations],
                         ['mapCurEve',self.updateMapCurEvePen],
                         ['mapPrevEve',self.updateMapPrevEvePen],
                         ['mapPolygon',self.updateMapPolygonPen],
                         ['archiveBackground',self.updateArchiveBackground],
                         ['archiveAvailability',self.updateArchiveAvailColor],
                         ['archiveSpanSelect',self.updateArchiveSpanColor],
                         ['archiveCurEve',self.updateArchiveCurEvePen],
                         ['archivePrevEve',self.updateArchivePrevEvePen]
                         ]:
            # If initializing Lazylyst or the value was edited, update GUI colors
            if init or self.pref['basePen'].val[key][3]:
                # Update the color
                func()
                # Reset the was-changed indicator
                self.pref['basePen'].val[key][3]=False
    
    # Update the text label colors
    def updateTextColor(self,init=False):
        col=QtGui.QColor(self.pref['basePen'].val['widgetText'][0])
        pen=mkPen(col)
        # Change all axis colors (axis line and ticks)
        for widget in self.staWidgets+[self.timeWidget,self.archiveEvent,
                                       self.archiveSpan,self.imageWidget]:
            widget.getPlotItem().getAxis('bottom').setPen(pen)
            widget.getPlotItem().getAxis('left').setPen(pen)
        self.mapWidget.setPen(pen)
        # Change the hovered station and time label colors
        self.mapWidget.hoverStaItem.setColor(col)
        self.archiveEvent.hoverTimeItem.setColor(col)
        self.archiveSpan.hoverTimeItem.setColor(col)
        # Change pick file title color
        title=self.timeWidget.getPlotItem().titleLabel.text
        self.timeWidget.getPlotItem().titleLabel.setText(title,color=col)  
        
    # Change the color of the trace background
    def updateTraceBackground(self,init=False):
        self.traceBgColors=self.getStaColors('traceBg')
        for aWidget in self.staWidgets:
            aWidget.setBackground(self.traceBgColors[aWidget.sta]) 

    # Change the color of time background
    def updateTimeBackground(self,init=False):
        col=QtGui.QColor(self.pref['basePen'].val['timeBackground'][0])
        self.timeWidget.setBackground(col)
        
    # Change the color of image background
    def updateImageBackground(self,init=False):
        col=QtGui.QColor(self.pref['basePen'].val['imageBackground'][0])
        self.imageWidget.setBackground(col)
    
    # Update the color of the map axis
    def updateMapBackground(self,init=False):
        col=QtGui.QColor(self.pref['basePen'].val['mapBackground'][0])
        self.mapWidget.setBackground(col)
    
    # Update the border to the image
    def updateImageBorderPen(self,init=False):
        col,width,depth=self.pref['basePen'].val['imageBorder'][0:3]
        self.imageWidget.imageBorder.setPen(mkPen(QtGui.QColor(col),width=width))
        self.imageWidget.imageBorder.setZValue(depth)
        if width==0:
            self.imageWidget.imageBorder.hide()
        else:
            self.imageWidget.imageBorder.show()
    
    # Update the color of all of the station spots, with user specified colors
    def updateMapStations(self,init=False):
        # Assign the station colors for the stations in staLoc
        self.mapStaColors=self.getStaColors('mapSta')
        staSize,staDep=self.pref['basePen'].val['mapStaDefault'][1:3]
        # Add the spots to the map
        self.mapWidget.loadStaLoc(self.hotVar['staLoc'].val,self.mapStaColors,
                                  staSize,staDep,init)
        
    # Update the polygon on the mapWidget
    def updateMapPolygon(self):
        # Remove old polygon
        self.mapWidget.removePolygon()
        # Add new if there are at least three vertices
        mapPoly=self.hotVar['mapPolygon'].val
        if len(mapPoly)>=3:
            self.mapWidget.loadPolygon(mapPoly)
            self.updateMapPolygonPen()
    
    # Update the pen of the map polygon
    def updateMapPolygonPen(self):
        self.mapWidget.updatePolygonPen(self.pref['basePen'].val['mapPolygon'][0:3])
    
    # Update the current event spots on the map widget
    def updateMapCurEve(self):
        self.mapWidget.loadEvePoints(self.hotVar['mapCurEve'].val,'cur')
        self.updateMapCurEvePen()
    
    # Update the previous event spots on the map widget
    def updateMapPrevEve(self):
        self.mapWidget.loadEvePoints(self.hotVar['mapPrevEve'].val,'prev')
        self.updateMapPrevEvePen()
    
    # Update the color of the current event symbols
    def updateMapCurEvePen(self): 
        self.mapWidget.updateEvePen(self.pref['basePen'].val['mapCurEve'][0:3],'cur')
        
    # Update the color of the previous event symbols
    def updateMapPrevEvePen(self):
        self.mapWidget.updateEvePen(self.pref['basePen'].val['mapPrevEve'][0:3],'prev')
    
    # Change the color of the archive axis
    def updateArchiveBackground(self):
        col=QtGui.QColor(self.pref['basePen'].val['archiveBackground'][0])
        self.archiveSpan.setBackground(col)
        self.archiveEvent.setBackground(col)
        
    # Change the color of the archive availability boxes
    def updateArchiveAvailColor(self):
        col=QtGui.QColor(self.pref['basePen'].val['archiveAvailability'][0])
        for box in self.archiveSpan.boxes:
            box.setPen(col)
            
    # Change the color of the archive span select
    def updateArchiveSpanColor(self):
        col=QtGui.QColor(self.pref['basePen'].val['archiveSpanSelect'][0])
        rgba=(col.red(),col.green(),col.blue(),65)
        self.archiveSpan.span.setBrush(mkBrush(rgba))
        for line in self.archiveSpan.span.lines:
            line.setPen(col)
        
    # Change the pen of the selected events in the archive visual, and the highlighted archive span color
    def updateArchiveCurEvePen(self):
        self.archiveEvent.updateEvePens(self.pref['basePen'].val['archiveCurEve'][0:3],'cur')
        # Use the select for highlighting the span when hovered
        col=QtGui.QColor(self.pref['basePen'].val['archiveCurEve'][0])
        for line in self.archiveSpan.span.lines:
            line.setHoverPen(col)    
            
    # Change the pen of the selected events in the archive visual, and the highlighted archive span color
    def updateArchivePrevEvePen(self):
        self.archiveEvent.updateEvePens(self.pref['basePen'].val['archivePrevEve'][0:3],'prev')
        
    # Update how many widgets are on the main page
    def updateStaPerPage(self,init=False):
        if not init:
            # Go back to the first page when updating number of staWidgets
            self.hotVar['curPage'].val=0
        # First remove the previous staWidgets
        for aWidget in self.staWidgets:
            aWidget.setParent(None)
            self.traceLayout.removeWidget(aWidget)
            self.staWidgets=[]
        axisPen=mkPen(QtGui.QColor(self.pref['basePen'].val['widgetText'][0]))
        # Add the desired number of staWidgets
        while len(self.staWidgets)<self.pref['staPerPage'].val:
            self.staWidgets.append(TraceWidget(self.mainLayout))
            self.staWidgets[-1].setXLink('timeAxis')
            self.traceLayout.addWidget(self.staWidgets[-1])
            # Connect the double click signal to the add pick signal
            self.staWidgets[-1].doubleClickSignal.connect(self.traceDoubleClickEvent)
            # Color the axes properly
            self.staWidgets[-1].getPlotItem().getAxis('left').setPen(axisPen)
        # If Lazylyst is initiating (first load), no need to update the page or cursor
        if not init:
            self.updatePage()
            self.updateCursor()
        
    # Set the archive availability
    def updateArchive(self,showBar=True,resetSearch=True):
        # Do nothing if directory not initialized
        if self.hotVar['archDir'].val=='':
            return
        # Load in all of the start times of the archive...
        archiveFiles,archiveTimes=getArchiveAvail(self.hotVar['archDir'].val,showBar=showBar)
        # ...and sort them by start time (required for when extracting an events data)
        argSort=np.argsort(archiveTimes[:,0])
        archiveFiles,archiveTimes=archiveFiles[argSort],archiveTimes[argSort]
        self.hotVar['archFiles'].val=archiveFiles
        self.hotVar['archFileTimes'].val=archiveTimes
        # Update the time boxes
        self.archiveSpan.updateBoxes(archiveTimes,self.pref['basePen'].val['archiveAvailability'][0])
        # Do not bother changing the span if nowhere to go
        if len(archiveFiles)==0:
            print('No miniseed files in '+self.hotVar['archDir'].val)
            return
        # Set the archive file search conditions with default
        if resetSearch:
            # Set the initial span boundaries, and x-limits
            t1,t2=np.min(archiveTimes[:,0]),np.max(archiveTimes[:,1])
            buff=(t2-t1)*0.05
            self.archiveSpan.pltItem.setXRange(t1-buff,t2+buff)
            self.archiveSpan.span.setRegion((t1,t2))
            # Also reset the archive list search text
            self.archiveListLineEdit.setText('')
        
    # Update the span text so user can set their bounds easier
    def updateSpanText(self):
        t0,t1=self.archiveSpan.getSpanBounds()
        self.archiveSpanT0Label.setText(str(t0)+'  to')
        self.archiveSpanT1Label.setText(' '+str(t1))
    
    # Update the pick list with events only in the selected span
    # ...this is done after both pickFiles and pickFileTimes are updated
    def updateArchiveSpanList(self):
        files=self.hotVar['pickFiles'].val
        times=self.hotVar['pickFileTimes'].val
        t1,t2=self.archiveSpan.span.getRegion()
        # See which files should be listed
        toListFiles=files[np.where((times>=t1)&(times<=t2))]
        # If the archive line edit is not empty, use to constrain search
        text=self.archiveListLineEdit.text()
        if text.replace(' ','')!='':
            toListFiles=[aFile for aFile in toListFiles if text in aFile]
        # Reload the pick file list
        self.archiveList.clear()
        self.archiveList.addItems(toListFiles)
        
    # Update the map with the new station metadata
    def updateStaMeta(self):
        # Load in the new station metadata
        if self.hotVar['staFile'].val.replace(' ','')=='':
            self.hotVar['staXml'].val=initHotVar()['staXml'].val
        else:
            self.hotVar['staXml'].val=readInventory(self.hotVar['staFile'].val)
        self.updateStaLoc()
        # Reset any additional map related visuals
        defaultHot=initHotVar()
        for key in ['mapCurEve','mapPrevEve','curMapSta','curMapPos']:
            self.hotVar[key].val=defaultHot[key].val
            self.hotVar[key].update()
    
    # Update staLoc along with the map widget
    def updateStaLoc(self,init=False):
        staLoc=staXml2Loc(self.hotVar['staXml'].val)
        # Project if desired
        if self.pref['staProjStyle'].val=='None' or init:
            self.hotVar['staLoc'].val=staLoc
        else:
            self.hotVar['staLoc'].val=projStaLoc(staLoc,self.pref['staProjStyle'].val)
        self.updateMapStations(init=True)
        
    # Update the map variables to be set when double clicking on it
    def updateMapDblClickedVars(self):
        self.hotVar['curMapSta'].val=self.mapWidget.selectSta
        self.hotVar['curMapPos'].val=self.mapWidget.clickPos
        
    # Load the pick file list for display given completly new pick directory
    def updatePickDir(self):
        # Reset the current pick file
        self.hotVar['curPickFile'].val=''
        self.hotVar['curPickFile'].update()
        self.refreshPickDirLists()
    
    # Update the pick file names and times, and their associated visuals
    def refreshPickDirLists(self):
        # Clear the old list
        self.archiveList.clear()
        self.hotVar['pickFiles'].val=self.getPickFiles()
        # Sort the pick files
        self.updateEveSort()
        # Update the archive event widget
        self.archiveEvent.updateEveLines(self.hotVar['pickFileTimes'].val,self.hotVar['curPickFile'].val)
        self.updateArchivePrevEvePen()
        self.updateArchiveCurEvePen()
                                         
    # Read correctly formatted pick files from the pick directory
    def getPickFiles(self):
        pickFiles,skipCount=[],0
        # Only accept files with proper naming convention
        for aFile in sorted(os.listdir(self.hotVar['pickDir'].val)):
            # Do not check folders
            if not os.path.isfile(self.hotVar['pickDir'].val+'/'+aFile):
                continue
            splitFile=aFile.split('_')
            # Make sure the file has the proper extension '.picks'
            if len(splitFile)!=2 or aFile.split('.')[-1]!='picks':
                skipCount+=1
                continue
            try:
                int(splitFile[0])
                getTimeFromFileName(aFile)
            except:
                skipCount+=1
                continue
            pickFiles.append(aFile)
        # Let user know if some files had incorrect file name convention
        if skipCount>0:
            print(str(skipCount)+' file(s) skipped when reading in the pickDir (convention: "ID_YYYYMMDD.HHMMSS.ffffff.picks")')
        return np.array(pickFiles,dtype=str)
    
    # With the same pick directory, force an update on the pick files...
    # ...this allows for addition of empty pick files and deletion of pick files
    def updatePickFiles(self):
        # Ensure that pick files returned are in wanted order
        self.updateEveSort()
        # Reset pick file time lines in archiveEvent
        self.archiveEvent.updateEveLines(self.hotVar['pickFileTimes'].val,self.hotVar['curPickFile'].val)
        self.updateArchivePrevEvePen()
        self.updateArchiveCurEvePen()
        # See which files were present prior to the update
        prevFiles=self.getPickFiles()
        # Delete events which are no longer present
        for aFile in prevFiles:
            path=self.hotVar['pickDir'].val+'/'+aFile
            if (aFile not in self.hotVar['pickFiles'].val) and os.path.exists(path):
                os.remove(path)
        # Add blank pick files which were not present before
        for aFile in self.hotVar['pickFiles'].val:
            if aFile not in prevFiles:
                newFile=open(self.hotVar['pickDir'].val+'/'+aFile,'w')
                newFile.close()
        # Update the current pick file if it is no longer in the pick files
        if self.hotVar['curPickFile'].val not in self.hotVar['pickFiles'].val:
            # Do not bother updating curPickFile if not set
            if self.hotVar['curPickFile'].val!='':
                self.hotVar['curPickFile'].val=''
                self.hotVar['curPickFile'].update()
    
    # Update the current pick file
    ## May want to add locking pick files ## (ex. use new variable curLockFile)
    ## How to deal with crashes? ##
    def updateCurPickFile(self):
        self.updateEvent()
            
    # Sort the pick files according to user preference...
    # ...also update the pickFileTimes
    def updateEveSort(self,init=False):
        # Sort alphabetically first (will be the secondary sorting)
        self.hotVar['pickFiles'].val=np.sort(self.hotVar['pickFiles'].val)
        self.hotVar['pickFileTimes'].val=np.array([getTimeFromFileName(aFile).timestamp for aFile in self.hotVar['pickFiles'].val])
        # If sorting by ID, this is the same as alphabetical (unless the ID is huge)
        if self.pref['eveSortStyle'].val=='id':
            pass
        elif self.pref['eveSortStyle'].val=='time':
            argSort=np.argsort(self.hotVar['pickFileTimes'].val)
            self.hotVar['pickFiles'].val=self.hotVar['pickFiles'].val[argSort]
            self.hotVar['pickFileTimes'].val=self.hotVar['pickFileTimes'].val[argSort]
        else:
            print('The eveSortStyle '+self.pref['eveSortStyle'].val+' has not been implemented, sorting by id')
        # Clear the old GUI list, and add the new items
        self.updateArchiveSpanList()
            
    # Update the image
    def updateImage(self):
        self.imageWidget.loadImage(self.hotVar['image'].val)
    
    # Update the strolling list with all in-use timed actions
    def updateStrollingList(self):
        strollers=self.qTimers.keys()
        # Update the label (header)
        self.strollingLabel.setText('Strolling ('+str(len(strollers))+')')
        # Reset the list
        self.strollComboBox.clear()
        self.strollComboBox.addItems(strollers)
        
    # Update the scheming list with all threads currently working
    def updateSchemingList(self):
        schemers=self.qThreads.keys()
        # Update the label (header)
        self.schemingLabel.setText('Scheming ('+str(len(schemers))+')')
        # Reset the list
        self.schemeComboBox.clear()
        self.schemeComboBox.addItems(schemers)
    
    # Set a span bound to user specified value
    def setSpanBoundViaDialog(self,whichBound):
        curBounds=self.archiveSpan.span.getRegion()
        if whichBound=='T0':
            bound=curBounds[0]
        else:
            bound=curBounds[1]
        newBound,ok=DateDialog.getDateTime(bound)
        # If the dialog was not cancelled...
        if not ok:
            return
        # Set the new bound
        elif whichBound=='T0':
            self.archiveSpan.span.setRegion((newBound,curBounds[1]))
        else:
            self.archiveSpan.span.setRegion((curBounds[0],newBound))
    
    # Update the cursor to be used on any graphics view widget (plot)
    def updateCursor(self,init=False):
        if self.pref['cursorStyle'].val=='arrow':
            cursor=QtGui.QCursor(QtCore.Qt.ArrowCursor)
        elif self.pref['cursorStyle'].val=='cross':
            cursor=QtGui.QCursor(QtCore.Qt.CrossCursor)
        for item in self.staWidgets+[self.mapWidget,self.imageWidget,
                                     self.archiveEvent,self.archiveSpan]:
            item.setCursor(cursor)
            
    # Take a screenshot and save to local directory
    def takeScreenshot(self):
        # Make the output directory if it does not exist
        aDir=self.hotVar['mainPath'].val+'/Screenshots'
        if not os.path.exists(aDir):
            os.makedirs(aDir)
        # Use the area bounding the main window to take an image of (docks outside will not be seen)
        screen=QtWidgets.QApplication.primaryScreen()
        pixMap=screen.grabWindow(0)
        if pixMap is not None:
            outName=time.strftime('%Y%m%d.%H%M%S.png',time.gmtime())
            print('Screenshot: '+outName)
            pixMap.save(aDir+'/'+outName, 'png')
        
    # Load setting from previous run, and initialize base variables
    def loadSettings(self):
        # Load the hot variables
        self.hotVar=initHotVar()
        # Get this scripts path, as functions will be relative to this
        mainPath=os.path.dirname(os.path.realpath(__file__))
        self.hotVar['mainPath'].val=mainPath
        for key,hotVar in iteritems(self.hotVar):
            hotVar.linkToFunction(self)
        # Get all values from settings
        self.setGen=QSettings(mainPath+'/setGen.ini', QSettings.IniFormat)
        self.setAct=QSettings(mainPath+'/setAct.ini', QSettings.IniFormat)
        self.setPref=QSettings(mainPath+'/setPref.ini', QSettings.IniFormat)
        self.setSource=QSettings(mainPath+'/setSource.ini', QSettings.IniFormat)
        # UI size
        if self.setGen.value("geometry") is not None:
            self.restoreGeometry(self.setGen.value("geometry"))
        if self.setGen.value("windowState") is not None:
            self.restoreState(self.setGen.value("windowState"))
        # Actions...
        self.act=self.setAct.value('actions', defaultActions())
        # ...reload locked actions from the defaults (may have been edited in a new version)
        for key,action in iteritems(defaultActions()):
            if action.locked:
                self.act[action.tag]=action
        # ...remove any locked actions which do no longer exist by default
        actKeys=self.act.keys()
        for key in actKeys:
            if self.act[key].locked and key not in defaultActions().keys():
                self.act.pop(key)
        # ...get the passive action ordering
        self.actPassiveOrder=self.setAct.value('actPassiveOrder', defaultPassiveOrder(self.act))
        if self.actPassiveOrder==None:
            self.actPassiveOrder=[]
        # ...link all actions to their appropriate functions and assign any missing attributes
        for key,action in iteritems(self.act):
            action.linkToFunction(self)
            action.fillMissingAttrib()
        # Preferences
        self.pref=defaultPreferences(self)
        prefVals=self.setPref.value('prefVals',{})
        for aKey in prefVals.keys():
            # Skip if this preference was removed in a newer version
            if aKey not in self.pref.keys():
                continue
            # Add default base color preferences not previously seen
            if aKey=='basePen':
                prefVals[aKey].update({oKey:self.pref[aKey].val[oKey] for oKey in self.pref[aKey].val.keys() if oKey not in prefVals[aKey].keys()})
            self.pref[aKey].val=prefVals[aKey]
        # Saved sources
        self.saveSource=self.setSource.value('savedSources',defaultSource())
        # Create empty variables
        self.staWidgets=[]
        self.qTimers={}
        self.qThreads={}
        self.traceSplitSizes=None
        self.setUserSeenAtTime()
        
    # Save all settings from current run
    def saveSettings(self,closing=True):
        # UI size and widget visibility
        self.setGen.setValue('geometry',self.saveGeometry())
        self.setGen.setValue('windowState',self.saveState())
        self.setGen.setValue('imageHidden',self.imageWidget.isHidden())
        # Actions, cannot save functions (will be linked again upon loading)
        for key in self.act.keys():
            self.act[key].func=None
        saveAct=deepcopy(self.act)
        self.setAct.setValue('actions',saveAct)
        self.setAct.setValue('actPassiveOrder',self.actPassiveOrder)
        # Preferences
        prefVals={}
        for aKey in self.pref.keys():
            prefVals[str(self.pref[aKey].tag)]=self.pref[aKey].val
        self.setPref.setValue('prefVals',prefVals)
        # Saved sources
        self.setSource.setValue('savedSources',self.saveSource)
        print('Settings saved')
        # If not closing point to the functions again
        if not closing:
            for key in self.act.keys():
                self.act[key].linkToFunction(self)
        
    # For actions which are triggered via built-in qt events
    def passAction(self):
        pass
    
    # When the GUI closes, will save to settings
    def closeEvent(self,ev):
        # Run any actions which are to be done before closing
        self.processAction(self.act['CloseLazylyst'])
        self.saveSettings()
        ev.accept()

# Class for logging
class QtHandler(logging.Handler):
    def __init__(self):
        logging.Handler.__init__(self)
    def emit(self, record):
        record = self.format(record)
        if record: XStream.stdout().write('%s\n'%record)

class XStream(QtCore.QObject):
    _stdout = None
    _stderr = None
    setTextCursor=QtCore.pyqtSignal()
    messageWritten=QtCore.pyqtSignal(str)
    def flush( self ):
        pass
    def fileno( self ):
        return -1
    def write( self, msg ):
        if ( not self.signalsBlocked() ):
            self.setTextCursor.emit()
            self.messageWritten.emit(str(msg))
    @staticmethod
    def stdout():
        if ( not XStream._stdout ):
            XStream._stdout = XStream()
            sys.stdout = XStream._stdout
        return XStream._stdout
    @staticmethod
    def stderr():
        if ( not XStream._stderr ):
            XStream._stderr = XStream()
            sys.stderr = XStream._stderr
        return XStream._stderr

def lazyExcepthook(aType, value, tback):
    # log the exception here
    ## FILL ##
    # then call the default handler
    sys.__excepthook__(aType, value, tback)
    
# Start up the logging and UI
def runLazylyst():
    # Allow errors to propogate without crashing
    sys.excepthook = lazyExcepthook
    # For sending stdout to the trace log
    logger = logging.getLogger(__name__)
    handler = QtHandler()
    handler.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    # Start up the UI
    app=QtWidgets.QApplication(sys.argv)
    window = LazylystMain()
    # Give an icon for the UI
    app.setWindowIcon(QtGui.QIcon(os.path.dirname(os.path.realpath(__file__))+'/Resources/LazyPenguin_64x64.png'))
    if os.name=='nt':
        myappid = u'yarp.lazylyst.'+__version__
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)
    window.show()
    sys.exit(app.exec_())

if __name__ == '__main__':
    runLazylyst()



from __future__ import print_function

import numpy as np

def streamFilter(*args,**kwargs):
    # If the returned trace is to have all channels seperated on the y-axis...
    # ...note for later
    if 'sepChas' in kwargs.keys():
        doSepChas=kwargs['sepChas']
        kwargs.pop('sepChas')
    else:
        doSepChas=False
    # If want to return the raw, or derivative...
    stream=args[0]
    if kwargs['type']=='raw':
        return stream
    try:
        stream.detrend()
    except:
        stream=stream.split()
        stream.detrend()
    if kwargs['type']=='derivative':
        stream.differentiate()
    else:
        # Process the remaining arguments as obspy filter would
        stream.filter(**kwargs)
    # If the channels were to be seperated, do after filtered
    if doSepChas:
        stream=sepChas(stream)
    return stream
    
# Shift traces up/down on a per-station basis so that they do not overlap
def sepChas(stream):
    # First sort the stream by channel name
    stream.sort(keys=['channel'])
    stas=np.array([tr.stats.network+'.'+tr.stats.station+'.'+tr.stats.location for tr in stream])
    for sta in np.unique(stas):
        args=np.where(stas==sta)[0]
        offset=0
        for arg in args:
            aMin,aMax=np.min(stream[arg].data),np.max(stream[arg].data)
            stream[arg].data-=(aMin-offset)
            offset+=aMax-aMin
    return stream

# Remove the instrument response
def invertResponse(stream,staXml):
    try:
        stream.remove_response(inventory=staXml, output='VEL')
    except:
        print('Could not remove instrument response')
    return stream
        
    


from __future__ import print_function
import os

import numpy as np
from obspy import UTCDateTime

# Function to change picking mode to the wanted mode...
# ... Return the original mode if the wanted mode is not defined  
def setPickMode(*args,**kwargs):
    availPickModes=sorted([str(key) for key in args[0].keys()])
    if kwargs['wantMode'] not in availPickModes:
        print('Update pickTypesMaxCountPerSta to be able to place '+str(kwargs['wantMode'])+' picks manually')
    return kwargs['wantMode']

# Swap between highlightint the vertical (P-picking mode) or horizontals (S-picking mode)
def setTracePenAssign(curMode,curAssign):
    # If highlight not defined, define it
    if 'highlight' not in curAssign.keys():
        curAssign={'highlight':['*Z'],'lowlight':['*1','*2','*E','*N']}
    # If pickMode is P, highlight vertical (also do this by default with no picking mode)
    if curMode in ['P','']:
        return {'highlight':['*Z'],'lowlight':['*1','*2','*E','*N']}
    # If pickMode is S, highlight the horizontal (toggle between horizontals if already highlighted)
    elif curMode=='S':
        if '*1' in curAssign['highlight']:
            return {'lowlight':['*1','*E','*Z'],'highlight':['*2','*N']}
        else:
            return {'lowlight':['*2','*N','*Z'],'highlight':['*1','*E']}
    # If not picking P or S, use default
    else:
        return {}
        
# Delete a given pick type, over a specific station (all stations if not specified)
def delPick(pickSet,pickMode,whichSta=None):
    if len(pickSet)==0:
        return '$pass'
    if whichSta is None:
        pickSet=pickSet[np.where((pickSet[:,1]!=pickMode))]
    else:
        pickSet=pickSet[np.where((pickSet[:,0]!=whichSta)|(pickSet[:,1]!=pickMode))]
    return pickSet
    
# Remove all picks from the current pick set
def delPickSet():
    return np.empty((0,3),dtype='a32')
        
# Set the current pick file
def setCurPickFile(curPickFile,pickFiles,nextFile=False,prevFile=False):
    maxFileIdx=len(pickFiles)-1
    prevPickFile=curPickFile
    # Return nothing if no pick files to choose from...
    # ...or if the current pick file is not set
    if maxFileIdx==-1 or curPickFile=='':
        return '$pass'
    curIdx=np.where(pickFiles==curPickFile)[0][0]
    # If the next or previous page, ensure still in bounds
    if nextFile and curIdx+1<=maxFileIdx:
        curPickFile=pickFiles[curIdx+1]
    elif prevFile and curIdx-1>=0:
        curPickFile=pickFiles[curIdx-1]
    # Do not update if nothing has changed
    if prevPickFile==curPickFile:
        return '$pass'
    return str(curPickFile)
    
# Remove the current pick file
def removeCurPickFile(curPickFile,pickFiles):
    if curPickFile in pickFiles:
        pickFiles=pickFiles[np.where(pickFiles!=curPickFile)]
    return pickFiles
    
# Go to the last double clicked station on the map 
def goToStaPage(curMapSta,staSort,staPerPage,curPage):
    if curMapSta is None:
        return '$pass'
    if curMapSta not in staSort:
        return '$pass'
    goToPage=np.where(staSort==curMapSta)[0][0]/staPerPage
    if goToPage==curPage:
        return '$pass'
    else:
        return int(goToPage)
        
# Pan between events
def goToNextPickFile(curPickFile,pickFiles):
    # If not pick files to go to, pass
    if len(pickFiles)==0:
        return '$pass'
    # If the current file isn't present, go to the first
    if curPickFile not in pickFiles:
        idx=0
    # Otherwise go to the next in the list
    else:
        idx=np.where((pickFiles==curPickFile))[0][0]+1
        if idx==len(pickFiles):
            idx=0
    return pickFiles[idx]
    
# Write out stream given current bounds as a MSEED file
# Gives name to file same as the pick file (different extension)
def writeMSEED(stream,timeRange,mainPath,curPickFile):
    outName=curPickFile.replace('.'+curPickFile.split('.')[-1],'.mseed')
    stream.trim(UTCDateTime(timeRange[0]),UTCDateTime(timeRange[1]))
    stream.write(mainPath+'/'+outName,format='MSEED')
    
# Add picks from nearby events with an alternate phase name (to color them differently)
def addNearbyPicks(pickDir,pickFiles,pickTimes,preTime,postTime
                   ,curPickFile,pickSet,givePickType='O'):
    # Get a list of all files whos data would overlap, given pre/post times
    curFileTime=pickTimes[list(pickFiles).index(curPickFile)]
    nearFiles=pickFiles[np.where((pickTimes+preTime<=curFileTime+postTime)&
                                 (pickTimes+postTime>=curFileTime+preTime))]
    # Add all of these files picks with type "givePickType"
    for aFile in nearFiles:
        path=pickDir+'/'+aFile
        # If the file is not empty and is not the current file
        if os.path.getsize(path)!=0 and aFile!=curPickFile:
            # Load the nearby picks
            addPicks=np.genfromtxt(path,delimiter=',',dtype='a32')
            if len(addPicks.shape)==1:
                addPicks=np.array([addPicks])
            # Swap the pick types and add to the current pick set
            addPicks[:,1]=givePickType
            pickSet=np.vstack((pickSet,addPicks))
    # If no picks present, nothing to change
    if 0 in pickSet.shape:
        return '$pass'
    # Remove any of the new picks which are outside the pre/post time
    if givePickType in pickSet[:,1]:
        pickSet=pickSet[np.where((pickSet[:,1]!=givePickType)|
                                 ((pickSet[:,2].astype(float)<=curFileTime+postTime)&
                                 (pickSet[:,2].astype(float)>=curFileTime+preTime)))]
        return pickSet
    # If no added picks, nothing to change
    else:
        return '$pass'
            
    
    
    
    
    
    
    


# Author: Andrew.M.G.Reynen
from __future__ import print_function
import importlib
import sys
from future.utils import iteritems

from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import Qt

from ActionSetup import Ui_actionDialog
from CustomFunctions import dict2Text, text2Dict

# Default Actions
def defaultActions():
    act={
    'OpenOptions':Action(tag='OpenOptions',name='openConfiguration',
                         path='$main',
                         trigger=Qt.Key_O,locked=True),
                      
    'ChangeSource':Action(tag='ChangeSource',name='openChangeSource',
                          path='$main',
                          trigger=Qt.Key_P,locked=True),

    'CloseLazylyst':Action(tag='CloseLazylyst',name='passAction',
                           path='$main',trigger='DoubleClick',locked=True),
                           
    'OpenLazylyst':Action(tag='OpenLazylyst',name='passAction',
                           path='$main',trigger='DoubleClick',locked=True),

    'ToggleImage':Action(tag='ToggleImage',name='toggleImageWidget',
                           path='$main',trigger=QtGui.QKeySequence('F4'),locked=True),

    'ReloadPlugins':Action(tag='ReloadPlugins',name='reloadPlugins',
                           path='$main',trigger=QtGui.QKeySequence('F5'),locked=True),

    'ReloadArchive':Action(tag='ReloadArchive',name='updateArchive',
                           path='$main',optionals={'showBar':False,'resetSearch':True},
                           trigger=QtGui.QKeySequence('F6'),threaded=True),
                           
    'SaveSettings':Action(tag='SaveSettings',name='saveSettings',
                          path='$main',optionals={'closing':False},
                          trigger=QtGui.QKeySequence('Ctrl+S'),locked=True),

    'Screenshot':Action(tag='Screenshot',name='takeScreenshot',
                        path='$main',trigger=QtGui.QKeySequence('F8'),locked=True),
                      
    'PageNext':Action(tag='PageNext',name='tabCurPage',
                      path='$main',optionals={'nextPage':True},
                      trigger=Qt.Key_D,locked=True),
                      
    'PagePrev':Action(tag='PagePrev',name='tabCurPage',
                      path='$main',optionals={'prevPage':True},
                      trigger=Qt.Key_A,locked=True),  
                      
    'FiltNone':Action(tag='FiltNone',name='streamFilter',
                      path='Plugins.Filters',optionals={'type':'raw'},
                      trigger=QtGui.QKeySequence('Shift+~'),inputs=['stream'],returns=['pltSt']),
    'FiltHP1':Action(tag='FiltHP1',name='streamFilter',
                      path='Plugins.Filters',optionals={'type':'highpass','freq':1,'corners':1,'zerophase':True},
                      trigger=QtGui.QKeySequence('Shift+Q'),inputs=['stream'],returns=['pltSt']),
    'FiltHP5':Action(tag='FiltHP5',name='streamFilter',
                      path='Plugins.Filters',optionals={'type':'highpass','freq':5,'corners':1,'zerophase':True},
                      trigger=QtGui.QKeySequence('Shift+W'),inputs=['stream'],returns=['pltSt']),
    'FiltHP10':Action(tag='FiltHP10',name='streamFilter',
                      path='Plugins.Filters',optionals={'type':'highpass','freq':10,'corners':1,'zerophase':True},
                      trigger=QtGui.QKeySequence('Shift+E'),inputs=['stream'],returns=['pltSt']),
    'FiltDeriv':Action(tag='FiltDeriv',name='streamFilter',
                      path='Plugins.Filters',optionals={'type':'derivative'},
                      trigger=QtGui.QKeySequence('Shift+R'),inputs=['stream'],returns=['pltSt']),

    'PickModeP':Action(tag='PickModeP',name='setPickMode',
                      path='Plugins.General',optionals={'wantMode':'P'},
                      trigger=QtGui.QKeySequence('1'),inputs=['pickTypesMaxCountPerSta'],returns=['pickMode']),

    'PickModeS':Action(tag='PickModeS',name='setPickMode',
                      path='Plugins.General',optionals={'wantMode':'S'},
                      trigger=QtGui.QKeySequence('2'),inputs=['pickTypesMaxCountPerSta'],returns=['pickMode']),

    'ToggleTracePen':Action(tag='ToggleTracePen',name='setTracePenAssign',
                            path='Plugins.General',passive=True,
                            trigger=['PickModeP','PickModeS','PickFileSetToClick','PickFileNext','PickFilePrev'],
                            inputs=['pickMode','tracePenAssign'],returns=['tracePenAssign']),

    'PickAdd':Action(tag='PickAdd',name='addClickPick',
                     path='$main',trigger='DoubleClick',locked=True),
                     
    'PickDelete':Action(tag='PickDelete',name='delPick',
                        path='Plugins.General',trigger=Qt.Key_4,
                        inputs=['pickSet','pickMode','curTraceSta'],returns=['pickSet']),

    'PickSetDelete':Action(tag='PickSetDelete',name='delPickSet',
                        path='Plugins.General',trigger=QtGui.QKeySequence('Shift+$'),
                        inputs=[],returns=['pickSet']),

    'PickFileCurDelete':Action(tag='PickFileCurDelete',name='removeCurPickFile',
                        path='Plugins.General',trigger=Qt.Key_Delete,
                        inputs=['curPickFile','pickFiles'],returns=['pickFiles']),
                     
    'PickFileSetToClick':Action(tag='PickFileSetToClick',name='setCurPickFileOnClick',
                                path='$main',trigger='DoubleClick',returns=['curPickFile'],locked=True),

    'PickFileNext':Action(tag='PickFileNext',name='setCurPickFile',path='Plugins.General',
                          optionals={'nextFile':True},trigger=QtGui.QKeySequence('Shift+D'),locked=True,
                          inputs=['curPickFile','pickFiles'],returns=['curPickFile']),

    'PickFilePrev':Action(tag='PickFilePrev',name='setCurPickFile',path='Plugins.General',
                          optionals={'prevFile':True},trigger=QtGui.QKeySequence('Shift+A'),locked=True,
                          inputs=['curPickFile','pickFiles'],returns=['curPickFile']),

    'SavePickSetOnNewEve':Action(tag='SavePickSetOnNewEve',name='savePickSet',
                                   path='$main',passive=True,
                                   trigger=['PickFileSetToClick','PickFileNext','PickFilePrev','CloseLazylyst'],
                                   beforeTrigger=True),
                                   
    'MapDblClicked':Action(tag='MapDblClicked',name='updateMapDblClickedVars',
                           path='$main',trigger='DoubleClick',locked=True),

    'GoToStaPage':Action(tag='GoToStaPage',name='goToStaPage',path='Plugins.General',
                         passive=True,trigger=['MapDblClicked'],
                         inputs=['curMapSta','staSort','staPerPage','curPage'],
                         returns=['curPage']),
                           
    'SimpleLocate':Action(tag='SimpleLocate',name='simpleLocator',path='Plugins.Locate',
                          passive=True,
                          trigger=['PickAdd','PickDelete','PickSetDelete','PickFileCurDelete',
                                   'PickFileNext','PickFilePrev','PickFileSetToClick','ChangeSource'],
                          inputs=['pickSet','staLoc','mapCurEve','staSort','sourceDict','staProjStyle'],
                          returns=['mapCurEve','traceBgPenAssign','mapStaPenAssign']),

    'StaSortAlph':Action(tag='StaSortAlph',name='staSortAlph',path='Plugins.Sorting',
                         trigger=QtGui.QKeySequence('Z'),
                         inputs=['staSort'],
                         returns=['staSort']),

    'StaSortPickTime':Action(tag='StaSortPickTime',name='staSortPickTime',path='Plugins.Sorting',
                         trigger=QtGui.QKeySequence('X'),
                         inputs=['staSort','pickSet','pickMode'],
                         returns=['staSort']),

    'StaSortDist':Action(tag='StaSortDist',name='staSortDist',path='Plugins.Sorting',
                         trigger=QtGui.QKeySequence('C'),
                         inputs=['staSort','staLoc','mapCurEve','staProjStyle'],
                         returns=['staSort']),

    'StaSortResidual':Action(tag='StaSortResidual',name='staSortResidual',path='Plugins.Sorting',
                         trigger=QtGui.QKeySequence('V'),
                         inputs=['staSort','staLoc','sourceDict','pickSet','mapCurEve','staProjStyle'],
                         returns=['staSort']),
                                  
    'StaSortReverse':Action(tag='StaSortReverse',name='staSortReverse',path='Plugins.Sorting',
                         trigger=QtGui.QKeySequence('B'),
                         inputs=['staSort'],
                         returns=['staSort']),
                              
    'ImageSpectVert':Action(tag='ImageSpectVert',name='spectrogramVert',path='Plugins.Image',
                            trigger=QtGui.QKeySequence('Ctrl+F'),
                            inputs=['pltSt','curTraceSta'],
                            returns=['image']),
                                                                    
    }
    return act
    
# Give the passive actions above some default ordering (ie. which is called first)
def defaultPassiveOrder(actions):
    order=sorted([act.tag for tag,act in iteritems(actions) if act.passive])
#    order=['SavePickSetOnNewEve','SimpleLocate'] ## Ensure this includes ALL default passive
    return order
    
# Capabilities of an action
class Action(object):
    def __init__(self,tag='New action',name='Function name',
                 path='Add path to function',optionals={},
                 passive=False,beforeTrigger=False,timer=False,
                 trigger='Add Trigger',inputs=[],returns=[],
                 timerInterval=60,func=None,locked=False,
                 sleeping=False,threaded=False):
        self.tag=tag # User visible name for the action
        self.name=name # Function name
        self.path=path # Function path (uses "." instead of "\", path is relative a directory in the systems PATH variable)
        self.optionals=optionals # Dictionary of the optional values which can be sent to the function
        self.passive=passive # If the action is passive, this is true
        self.beforeTrigger=beforeTrigger # If the passive function should be applied before/after the function
        self.timer=timer # If active, can use this to use a QTimer to set things off
        self.timerInterval=timerInterval # Interval time in seconds for the timer
        self.trigger=trigger # Trigger of the action
        self.inputs=inputs # Hot variables to be sent as inputs to the function (in the correct order)
        self.returns=returns # Hot variables to be returned for update
        self.func=func # Function which is called upon trigger
        self.locked=locked # If the action is allowed to be altered (other than the trigger)
        self.sleeping=sleeping # If the action is responding to triggers or not
        self.threaded=threaded # If the action is to be run on a seperate thread ("in background")
        # convert the trigger to a key sequence, if the trigger was a key
        if type(trigger)==type(Qt.Key_1):
            self.trigger=QtGui.QKeySequence(self.trigger)
    
    # When loading the previous settings, must link the action to its functions again
    def linkToFunction(self,main,reloadMod=False):
        # If the path does not relate to already defined function locations
        if self.path not in ['$main']:
            # Extract the function
            try:
                mod=importlib.import_module(self.path)
                # Force reloading of the module if wanted
                if reloadMod:
                    if sys.version_info[0]==2:
                        reload(mod)
                    else:
                        importlib.reload(mod)
                func=getattr(mod,self.name)
            except Exception as error:
                print(error)
                print('Action '+self.tag+' did not load from '+self.path+'.'+self.name)
                return
        # Otherwise, grab function from predefined location
        else:
            try:
                if self.path=='$main':
                    func=getattr(main,self.name)
            except:
                print(self.tag+' did not load from '+self.path+'.'+self.name)
                return
        # Assign the function to the action
        self.func=func
        return
    
    # Assign the default attributes which were are added in newer versions...
    # ...if not already present
    def fillMissingAttrib(self):
        defAct=Action()
        for attrib in ['sleeping','threaded']:
            try:
                getattr(self,attrib)
            except:
                setattr(self,attrib,getattr(defAct,attrib))
        
# Action setup dialog
class ActionSetupDialog(QtWidgets.QDialog, Ui_actionDialog):
    def __init__(self,main,action,actDict,hotVar,pref,tempLock,parent=None):
        QtWidgets.QDialog.__init__(self,parent)
        self.setupUi(self)
        self.trigReminder=False # Used for reminder to user if toggling between active/passive
        self.main=main # The main window
        self.action=action # The action to update
        self.actDict=actDict # All other actions
        self.hotVar=hotVar # Inputs/returns to go to/from function
        self.pref=pref # Additional inputs to go to the function
        # Give the dialog some functionaly
        self.setFunctionality()
        # Load in the text to the related action
        self.fillDialog()
        # Disable the majority of this gui if the action is locked...
        if self.action.locked or tempLock:
            self.lockDialog()
        # ...also, if this is only click event (double click for adding one pick)
        if self.action.trigger=='DoubleClick':
            self.actTriggerLineEdit.setEnabled(False)
     
    # Set up some functionality to the action set up dialog
    def setFunctionality(self):
        # For radio buttons, passive/active
        self.actActiveRadio.clicked.connect(lambda: self.togglePassiveActive())
        self.actPassiveRadio.clicked.connect(lambda: self.togglePassiveActive())
        # For the tag line edit, to remind if a tag is already being used
        self.actTagLineEdit.hoverOut.connect(self.unqTagName)
        # For the trigger line (active) and lists (passive)
        self.actTriggerLineEdit.keyPressed.connect(self.updateKeyBind)
        self.actAvailTriggerList.itemDoubleClicked.connect(lambda: self.addAvailVar('trigger'))
        self.actSelectTriggerList.keyPressedSignal.connect(lambda: self.removeSelectVar('trigger'))
        self.actSelectTriggerList.itemDoubleClicked.connect(lambda: self.removeSelectVar('trigger'))
        # For the input/return, avail/select lists
        self.actAvailInputList.itemDoubleClicked.connect(lambda: self.addAvailVar('input'))
        self.actAvailReturnList.itemDoubleClicked.connect(lambda: self.addAvailVar('return'))
        self.actSelectInputList.keyPressedSignal.connect(lambda: self.removeSelectVar('input'))
        self.actSelectReturnList.keyPressedSignal.connect(lambda: self.removeSelectVar('return'))
        self.actSelectInputList.itemDoubleClicked.connect(lambda: self.removeSelectVar('input'))
        self.actSelectReturnList.itemDoubleClicked.connect(lambda: self.removeSelectVar('return'))
    
    # Fill the dialog with info relating to action
    def fillDialog(self):
        # Fill in the text...
        # ...line edit entries
        self.actTagLineEdit.setText(self.action.tag)
        self.actNameLineEdit.setText(self.action.name)
        self.actPathLineEdit.setText(self.action.path)
        # ...optionals line edit
        self.actOptionalsLineEdit.setText(dict2Text(self.action.optionals))
        # ...trigger line edit entry
        if self.action.passive or self.action.trigger=='Set Trigger':
            self.actTriggerLineEdit.setText('Set Trigger')
        elif self.action.trigger=='DoubleClick':
            self.actTriggerLineEdit.setText(self.action.trigger)
        else:
            self.actTriggerLineEdit.setText(self.action.trigger.toString())
        # ...trigger list, use only triggers which are active
        self.passiveTriggers=[action.tag for key,action in iteritems(self.actDict) if not action.passive]
        self.actAvailTriggerList.addItems(sorted(self.passiveTriggers))
        # Set the appropriate radio button on
        if self.action.passive:
            self.actPassiveRadio.setChecked(True)
        else:
            self.actActiveRadio.setChecked(True)
        # Set the state of the beforeTrigger check box (used for passive only)
        if self.action.passive and self.action.beforeTrigger:
            self.passiveBeforeCheck.setChecked(True)
        else:
            self.passiveBeforeCheck.setChecked(False)
        # Set the state of the activeTimer check box (used for active only)
        if not self.action.passive and self.action.timer:
            self.activeTimerCheck.setChecked(True)
        else:
            self.activeTimerCheck.setChecked(False)
        # Set the state of the activeThreaded check box
        if self.action.threaded:
            self.activeThreadedCheck.setChecked(True)
        else:
            self.activeThreadedCheck.setChecked(False)
        # Set the timer interval value
        self.actIntervalLineEdit.setText(str(self.action.timerInterval))
        # Toggle the appropriate radio button
        self.togglePassiveActive(init=True)
        # Fill in the available inputs and returns...
        # ...hot variables can go into inputs and outputs
        self.addTipItems(self.actAvailInputList,[hotVar for key,hotVar in iteritems(self.hotVar)])
        self.addTipItems(self.actAvailReturnList,[hotVar for key,hotVar in iteritems(self.hotVar) if hotVar.returnable])
        # ...preferences are only allowed as inputs (no need to add colors here, just adds clutter)
        self.addTipItems(self.actAvailInputList,[pref for key,pref in iteritems(self.pref) if not 
                                                 (key in ['basePen','customPen','pickPen','cursorStyle'])])
        # Fill in the selected triggers (passive),inputs and returns
        if self.action.passive:
            self.actSelectTriggerList.addItems(self.action.trigger)
        self.addTipItems(self.actSelectInputList,self.action.inputs,useKey=True)
        self.addTipItems(self.actSelectReturnList,self.action.returns,useKey=True)
    
    # Add hot variable or preference to a designated list widget
    def addTipItems(self,listWidget,tipObjects,useKey=False):
        for tipObject in tipObjects:
            # If the key was passed, instead of the object - get the object
            if useKey:
                if tipObject in self.pref.keys():
                    tipObject=self.pref[tipObject]
                else:
                    tipObject=self.hotVar[tipObject]
            item=QtWidgets.QListWidgetItem()
            item.setText(tipObject.tag)
            item.setToolTip(tipObject.tip)
            listWidget.addItem(item)
    
    # Lock the dialog so that values cannot be updated, for some built-in actions
    def lockDialog(self):
        for widget in [self.actNameLineEdit,self.actPathLineEdit,
                       self.actOptionalsLineEdit,self.actAvailTriggerList,
                       self.actSelectTriggerList,self.actIntervalLineEdit,
                       self.actAvailInputList,self.actAvailReturnList,
                       self.actSelectInputList,self.actSelectReturnList,
                       self.actPassiveRadio,self.actActiveRadio,
                       self.passiveBeforeCheck,self.activeTimerCheck,
                       self.activeThreadedCheck,self.actTagLineEdit]:
            widget.setEnabled(False)
    
    # Update the key bind to what the user pressed
    def updateKeyBind(self,keyBindText):
        # Key binds are only used for active actions
        if self.actPassiveRadio.isChecked():
            return
        # If the keybind is already in use, do not allow update
        for tag,action in iteritems(self.actDict):
            # Only active actions have keybinds (do not need to check passive)
            if action.passive or action.trigger in ['DoubleClick','Set Trigger']:
                continue
            if action.trigger.toString()==keyBindText and action.tag!=self.action.tag:
                print(action.tag+' already uses Keybind '+keyBindText)
                return
        # Otherwise, update the trigger line edit (trigger value is updated upon close)
        self.actTriggerLineEdit.setText(keyBindText)
    
    # Move the user chosen available hot variable or preference to the selected list
    def addAvailVar(self,listTag):
        # See which pair of lists was called
        if listTag=='return':
            fromList=self.actAvailReturnList
            toList=self.actSelectReturnList
        elif listTag=='input':
            fromList=self.actAvailInputList
            toList=self.actSelectInputList
        else:
            fromList=self.actAvailTriggerList
            toList=self.actSelectTriggerList
        # Add to the selected list, if the item is not already there
        aKey=fromList.currentItem().text()
        if aKey not in [toList.item(i).text() for i in range(toList.count())]:
            # If this was from the return/input, add the tag to the item
            if listTag in ['return','input']:
                self.addTipItems(toList,[aKey],useKey=True)
            else:
                toList.addItem(aKey)
    
    # Removes the current item from a selected list of hot variables and/or preferences
    def removeSelectVar(self,listTag):
        # See which list was selected
        potTags=['return','input','trigger']
        potLists=[self.actSelectReturnList,self.actSelectInputList,self.actSelectTriggerList]
        aList=potLists[potTags.index(listTag)]
        # Remove the item if one is selected
        if aList.currentItem!=None:
            if aList.currentItem().isSelected():
                aList.takeItem(aList.currentRow())
            
    # Depending on if this is an active or passive action, turn on/off some widgets
    def togglePassiveActive(self,init=False):
        active=self.actActiveRadio.isChecked()
        self.actTriggerLineEdit.setEnabled(active)
        if active:
            self.actTriggerLineEdit.setReadOnly(active)
        self.actIntervalLineEdit.setEnabled(active)
        self.activeTimerCheck.setEnabled(active)
        self.activeThreadedCheck.setEnabled(active)
        self.passiveBeforeCheck.setEnabled(not active)
        self.actAvailTriggerList.setEnabled(not active)
        self.actSelectTriggerList.setEnabled(not active)
        # If the user has change from passive to active, remind that this will not be saved...
        # ...unless the trigger value is updated
        if (not self.trigReminder) and (not init):
            print('If changing between active and passive, update the trigger value')
            self.trigReminder=True

    # Check the tag name to ensure that it doesn't conflict with other actions
    def unqTagName(self):
        curTag=self.actTagLineEdit.text()
        for tag,action in iteritems(self.actDict):
            if curTag==tag and action.tag!=self.action.tag:
                print('This tag is already in use - if left as is changes will not be accepted')
                return False
        return True
        
    # Check that the trigger is appropriate relative to the action type
    def appropTrigger(self):
        # If a passive action, all triggers should be in the trigger list
        if self.actPassiveRadio.isChecked():
            sendFalse=False
            for entry in self.actSelectTriggerList.visualListOrder():
                if entry not in self.passiveTriggers:
                    print('The active trigger tag '+entry+' no longer exists')
                    sendFalse=True
            if sendFalse:
                return False
        # If a active action, the trigger should be able to convert to a keybind
        else:
            if self.actTriggerLineEdit.text()=='Set Trigger':
                return False
            try:
                QtGui.QKeySequence(self.actTriggerLineEdit.text())
            except:
                return False
        return True
    
    # Upon close, fill in the action with all of the selected information and return
    def returnAction(self):
        # First check to see that the new parameters make sense and the action is to be updated,
        # ... checking that the tag and the trigger are appropriate, also that the timer makes sense
        if self.action.trigger=='DoubleClick':
            print('No changes can be made to actions with DoubleClick triggers')
            return None
        if self.actTagLineEdit.text()=='New action':
            print('Action update declined, tag was still default')
            return None
        if not self.unqTagName():
            print('Action update declined, tag is not unique')
            return None
        if not self.appropTrigger():
            print('Action update declined, trigger was not set correctly')
            return None
        if self.activeTimerCheck.isChecked():
            try:
                float(self.actIntervalLineEdit.text())
            except:
                if float(self.actIntervalLineEdit.text())>0:
                    pass
                print('Action update declined, timer interval was not a positive number')
                return None
        # Set information from line edits...
        self.action.tag=self.actTagLineEdit.text()
        self.action.name=self.actNameLineEdit.text()
        self.action.path=self.actPathLineEdit.text()
        # ...optionals, set the values to the accepted format type (in order: bool,float,int,str)
        optText=self.actOptionalsLineEdit.text()
        try:
            optionals=text2Dict(optText)
        except:
            print('optionals were not formatted correctly, leaving as blank')
            optionals={}
        self.action.optionals=optionals
        # ...set whether this is an active or passive action
        self.action.passive=self.actPassiveRadio.isChecked()
        # ...trigger value
        if self.action.passive:
            self.action.trigger=self.actSelectTriggerList.visualListOrder()
        else:
            self.action.trigger=QtGui.QKeySequence(self.actTriggerLineEdit.text())
        # Set info from radio buttons (passive/active) 
        self.action.passive=self.actPassiveRadio.isChecked()
        # If the action should be applied before the trigger (used for passive only)
        if self.passiveBeforeCheck.isChecked():
            self.action.beforeTrigger=True
        # If the action should set off with a given interval
        if self.activeTimerCheck.isChecked():
            self.action.timer=True
            self.action.timerInterval=float(self.actIntervalLineEdit.text())
        else:
            self.action.timer=False
        # If the action should initiate a seperate thread when triggered
        if self.activeThreadedCheck.isChecked():
            self.action.threaded=True
        else:
            self.action.threaded=False
        # Collect the tags of the inputs and returns associated with the action
        self.action.inputs=self.actSelectInputList.visualListOrder()
        self.action.returns=self.actSelectReturnList.visualListOrder()
        # Try to link to the function (given the new path, and name)
        try:
            self.action.linkToFunction(self.main)
        except:
            print('Failed to link '+self.action.tag+' to '+self.action.name+' at '+self.action.path)
        return self.action
        
# Custom thread class to execute a queue of actions
class QueueThread(QtCore.QThread):
    setNextInputs=QtCore.pyqtSignal()    
    sendReturns=QtCore.pyqtSignal()
    
    def __init__(self,tag,actQueue):
        QtCore.QThread.__init__(self)
        self.tag=tag # The tag of the active action which created this queue
        self.actQueue=actQueue # The queue of all actions relating to the active action
        self.inputs=None # The inputs of the action currently being processed
        self.returns=None # The returns of the last action processed
        self.curIdx=None # Which index within the action queue is being processed
    
    def setInputs(self,inputs):
        self.inputs=inputs
        
    def resetInputsAndReturns(self):
        self.returns=None
        self.inputs=None
    
    # This function is called using the start() function
    def run(self):
        # Go through each of the queued action and execute it
        for i in range(len(self.actQueue)):
            # Ask to collect the required inputs
            self.curIdx=i
            self.setNextInputs.emit()
            # Wait to collect the inputs before processing the action
            while self.inputs==None:
                self.msleep(50)
            # Run the actions function and send back the return values
            self.returns=self.actQueue[self.curIdx].func(*self.inputs,**self.actQueue[self.curIdx].optionals)
            self.sendReturns.emit()
            # Wait for the return values on the GUI to be updated
            while self.returns is not None:
                self.msleep(50)
        self.exit()



# Author: Andrew.M.G.Reynen
from __future__ import print_function
import importlib
import os
import sys
from copy import deepcopy
from future.utils import iteritems
if sys.version_info.major == 3:
    unicode = str
    
import numpy as np
from obspy import Stream, read_inventory
from obspy.core.inventory import Inventory

from CustomFunctions import getTimeFromFileName,getStaStr

# Default Hot Variables
def initHotVar():
    hotVar={
    'stream':HotVar(tag='stream',val=Stream(),dataType=type(Stream()),returnable=False,
                    tip='Raw trace data from archive for current event'),
    'pltSt':HotVar(tag='pltSt',val=Stream(),dataType=type(Stream()),
                   funcName='updateTraces',checkName='checkPltSt',
                    tip='Plotted trace data for current event'),
    'staSort':HotVar(tag='staSort',val=[],dataType=type(np.array([''])),
                     funcName='updatePage',checkName='checkStaSort',
                     tip='Station codes, order defines trace data widget ordering'),
    'curPage':HotVar(tag='curPage',val=0,dataType=int,
                     funcName='updateCurPage',
                     tip='Current page index'),
    'timeRange':HotVar(tag='timeRange',val=[0,1],dataType=list,
                       funcName='updateTimeRange',checkName='checkTimeRange',
                       tip='Time limits on the trace time widget'),
    'yTraceRanges':HotVar(tag='yTraceRanges',val=np.empty((0,2)),dataType=type(np.array([0.0])),
                       funcName='updateTraceRangesY',checkName='checkTraceRangesY',
                       tip='Y-limits on the trace data widgets'),
    'sourceTag':HotVar(tag='sourceTag',val='',dataType=str,
                       funcName='updateSource',checkName='checkSourceTag',
                       tip='Tag given to the current source'),
    'pickDir':HotVar(tag='pickDir',val='',dataType=str,
                     funcName='updatePickDir', checkName='checkPickDir',
                     tip='Pick directory path'),
    'pickFiles':HotVar(tag='pickFiles',val=[],dataType=type(np.array(['str'])), 
                       funcName='updatePickFiles',checkName='checkPickFileNames',
                       tip='File names contained within the pick directory'),
    'pickFileTimes':HotVar(tag='pickFileTimes',val=[],dataType=type(np.array([0.0])),returnable=False,
                           tip='Timestamps contained in the pick file names'),   
    'curPickFile':HotVar(tag='curPickFile',val='',dataType=str,
                         funcName='updateCurPickFile',checkName='checkCurPickFile',
                         tip='Filename of the currently selected pick file'),
    'pickSet':HotVar(tag='pickSet',val=np.empty((0,3)),dataType=type(np.array([0.0])),
                     funcName='updatePagePicks',checkName='checkPickSet',
                     tip='Picks assigned to the current pick file'),
    'pickMode':HotVar(tag='pickMode',val='',dataType=str,
                      tip='Current pick phase in use'),
    'tracePenAssign':HotVar(tag='tracePenAssign',val={},dataType=dict,
                            funcName='updateTracePen',checkName='checkTracePenAssign',
                            tip='Pen values assigned to traces data on a channel basis'),
    'traceBgPenAssign':HotVar(tag='traceBgPenAssign',val={},dataType=dict,
                            funcName='updateTraceBackground',checkName='checkStaColAssign',
                            tip='Colors assigned to trace data widget backgrounds, station basis'),
    'mapStaPenAssign':HotVar(tag='mapStaPenAssign',val={},dataType=dict,
                            funcName='updateMapStations',checkName='checkStaColAssign',
                            tip='Colors assigned to stations'),
    'mapCurEve':HotVar(tag='mapCurEve',val=np.empty((0,5)),dataType=type(np.array([0.0])),
                       funcName='updateMapCurEve',checkName='checkEveArr',
                       tip='Events plotted in the foreground of the map widget'),
    'mapPrevEve':HotVar(tag='mapPrevEve',val=np.empty((0,5)),dataType=type(np.array([0.0])),
                       funcName='updateMapPrevEve',checkName='checkEveArr',
                       tip='Events plotted in the background of the map widget'),
    'mapPolygon':HotVar(tag='mapPolygon',val=np.empty((0,2),dtype=float),dataType=type(np.array([0.0])),
                       funcName='updateMapPolygon',checkName='checkMapPolygon',
                       tip='Vertices of the polygon on the map widget'),
    'archDir':HotVar(tag='archDir',val='',dataType=str,
                     funcName='updateArchive',checkName='checkArchDir',
                     tip='Archive directory path'),
    'archFiles':HotVar(tag='archFiles',val=[],dataType=type(np.array([0.0])),returnable=False,
                       tip='File paths contained within the archive directory'),
    'archFileTimes':HotVar(tag='archFileTimes',val=[],dataType=type(np.array([0.0])),returnable=False,
                           tip='Earliest and latest timestamp contained in the archive files'),   
    'curTraceSta':HotVar(tag='curTraceSta',val='',dataType=str,returnable=False,
                         tip='Network.Station.Location of the currently hovered over trace data widget'),
    'curTracePos':HotVar(tag='curTracePos',val=[0,0],dataType=list,returnable=False,
                         tip='Mouse position [X,Y] of the currently hovered over trace data widget'),
    'staFile':HotVar(tag='staFile',val='',dataType=str,
                     funcName='updateStaMeta',checkName='checkStaFile',
                     tip='Station file path'),
    'staLoc':HotVar(tag='staLoc',val=np.empty((0,4)),dataType=type(np.array([0.0])),returnable=False,
                     tip='Station locations'),
    'staXml':HotVar(tag='staXml',val=Inventory(networks=[],source='Lazylyst'),
                    dataType=type(Inventory(networks=[],source='Lazylyst')),returnable=False,
                     tip='Station metadata'),
    'curMapSta':HotVar(tag='curMapSta',val='',dataType=str,returnable=False,
                       tip='Last double clicked station on the map widget'),
    'curMapPos':HotVar(tag='curMapPos',val=[0,0],dataType=list,returnable=False,
                       tip='Last double clicked position on the map widget'),
    'mainPath':HotVar(tag='mainPath',val='',dataType=str,returnable=False,
                      tip='Path which contains Lazylst'),
    'image':HotVar(tag='image',val={'data':np.zeros((1,1)),'tDelta':1,'t0':0},dataType=dict,
                   funcName='updateImage',checkName='checkImage',
                   tip='Currently plotted image on the image widget'),
    'customDict':HotVar(tag='customDict',val={},dataType=dict,
                        tip='Dictionary to hold arbitrary variables'),
    'sourceDict':HotVar(tag='sourceDict',val={},dataType=dict,
                        checkName='checkSourceDict',
                        tip='Dictionary to hold source specific variables'),
    'afkTime':HotVar(tag='afkTime',val=0,dataType=float,returnable=False,
                        tip='Time in seconds since last user interaction'),
    }
    return hotVar

# Class for hot variables, which have defined update functions
class HotVar(object):
    def __init__(self,tag=None,val=None,
                 dataType=None,func=None,
                 funcName=None,returnable=True,
                 check=None,checkName=None,tip=''):
        self.tag=tag
        self.val=val
        self.dataType=dataType
        self.func=func # Function which is called upon update
        self.funcName=funcName # Name of the function within $main to be called
        self.returnable=returnable # If this hot variable is allowed to be returned for update
        self.check=check # Function used to check the validity of the input data
        self.checkName=checkName # Name of the check function within $main
        self.tip=tip # Short description of the hot variable
    
    # Return a deep copy of the objects value
    def getVal(self):
        # If a list has no entries, return the default (empty list) instead of the deepcopy
        if self.dataType in [list,type(np.array([0.0]))]:
            if len(self.val)==0:
                return initHotVar()[self.tag].val
        # The station xml file, and customDict can take a while to copy - user will be able to edit
        if self.tag in ['staXml','customDict']:
            return self.val
        elif self.tag=='stream':
            return self.val.copy()
        else:
            return deepcopy(self.val)
        
    # Call the specified function to update the hot variable value
    def update(self):
        if self.func is None:
            return
        self.func()
    
    # Link a hot variable to its pre-defined update and check functions
    def linkToFunction(self,main):
        # If there is no update function, skip
        if self.funcName is None and self.checkName is None:
            return
        # Link to the check function first (if present)
        if self.checkName is None:
            linkCheck=True
        else:
            try:
                self.check=getattr(importlib.import_module('HotVariables'),self.checkName)
                linkCheck=True
            except:
                linkCheck=False
                print(self.tag+' check function did not load from $main.'+self.checkName)
        # If the link to the check function passed, link to the update function (if present)
        if linkCheck and self.funcName is not None:
            try:
                self.func=getattr(main,self.funcName)
            except:
                print(self.tag+' update function did not load from $main.'+self.funcName)

# Ensure that the new plot stream has the same combination of stations as stream
def checkPltSt(main,pltSt):
    oStas=np.unique([getStaStr(tr) for tr in main.hotVar['stream'].val])
    nStas=np.unique([getStaStr(tr) for tr in pltSt])
    if not np.array_equal(np.sort(oStas),np.sort(nStas)):
        print('The return pltSt does not have all and only the stations (N.S.L) present in stream')
        return False
    return True
    
# Ensure that the new station sorting doesn't have stations for which there are no traces
def checkStaSort(main,newSort):
    trStas=np.unique([getStaStr(tr) for tr in main.hotVar['pltSt'].val])
    if not np.array_equal(np.sort(trStas),np.sort(newSort)):
        print('The return staSort does not have all and only the stations (N.S.L) present in pltSt')
        return False
    return True
    
# Ensure that the limits are in the proper order, can be floats, and just contains two values
def checkTimeRange(main,timeRange):
    try:
        np.array(timeRange,dtype=float)
    except:
        print('timeRange values must be numbers (timestamp (s))')
        return False
    if len(timeRange)!=2:
        print('For timeRange, two values are required to specify the time range, got '+str(len(timeRange)))
        return False
    elif timeRange[0]>=timeRange[1]:
        print('For timeRange, the second (right) limit must be greater than the first limit')
        return False
    return True

# Ensure that limits are in proper order, can be floats, and have as many entries as there are staWidgets
def checkTraceRangesY(main,yRanges):
    try:
        yRanges=np.array(yRanges,dtype=float)
    except:
        print('yTraceRanges values must be numbers')
        return False
    if len(yRanges.shape)!=2 or len(yRanges)!=len(main.staWidgets):
        print('yTraceRanges must be two dimensional, and must have length=staPerPage('+
              str(main.pref['staPerPage'].val)+')')
        return False
    elif np.sum(np.diff(yRanges,axis=1)<=0)!=0:
        print('For yTraceRanges, limits for all trace widgets must increase ie. [low,high]')
        return False
    return True

# Ensure that the specified tag actually exists
def checkSourceTag(main,tag):
    if tag not in main.saveSource.keys():
        print('The sourceTag '+tag+' is not currently a saved source')
        return False
    return True

# Ensure that the source dictionary contains only dictionaries, and its values are accepted
def checkSourceDict(main,sourceDict):
    for groupName in sourceDict.keys():
        # Entries must also be dictionaries
        if type(sourceDict[groupName])!=dict:
            print('The sourceDict had a group which was not a dictionary')
            return False
        for varName,varVal in iteritems(sourceDict[groupName]):
            # Check that value is in supported types
            if type(varVal) not in [str,int,float,type(np.array([]))]:
                print('The sourceDict had a value "'+str(varName)+'" with type not in [str,int,float,numpy.array]')
                return False
            # If it is a numpy array, check that it is less than 2 dimensions
            elif type(varVal)==type(np.array([])):
                if len(varVal.shape)>1:
                    print('sourceDict currently only support numpy arrays that are 0 or 1 dimensional')
                    return False
    return True

# Ensure that the supplied pick directory actually exists (make one if it does not)
def checkPickDir(main,pickDir):
    if not os.path.exists(pickDir):
        try:
            print('The pickDir did not exist, folder has been created')
            os.makedirs(pickDir)
        except:
            print('The pickDir did not exist, supplied string not compatible as a directory name')
            return False
    return True

# Ensure that all of the returned pick files conform to the proper name
def checkPickFileNames(main,pickFiles):
    passTest=True
    for i,aFile in enumerate(pickFiles):
        splitFile=aFile.split('_')
        if len(splitFile)!=2 or 'picks'!=aFile.split('.')[-1]:
            print(aFile+' does not match format IntegerID_%Y%m%d.%H%M%S.%f.picks')
            passTest=False
            continue
        try:
            getTimeFromFileName(aFile)
            int(splitFile[0])
            continue
        except:
            print(aFile+' does not match format IntegerID_%Y%m%d.%H%M%S.%f.picks')
            passTest=False
    if len(np.unique(pickFiles))!=len(pickFiles):
        print('Pick file names were non unique')
        passTest=False
    return passTest

# Ensure that the new pick file has the proper naming convention
def checkCurPickFile(main,pickFile):
    # If the pick directory has not been set, don't try
    if main.hotVar['pickDir'].val=='':
        print('The pickDir has not been set')
        return False
    # If the user wants to return to a blank screen, let them
    if pickFile=='':
        return True
    # Check to ensure name convention is maintained
    if not checkPickFileNames(main,[pickFile]):
        return False
    # If the pick file exists and is present in pickFiles, allow...
    if (pickFile in main.hotVar['pickFiles'].val and 
        os.path.exists(main.hotVar['pickDir'].val+'/'+pickFile)):
        return True
    # ...otherwise check to ensure that the directory and pick files variable are synced
    elif set(main.hotVar['pickFiles'].val)!=set(os.listdir(main.hotVar['pickDir'].val)):
        print('Hot variable pickDir is not synced with pickFiles, refreshing')
        main.refreshPickDirLists()
    # After refreshing the current pick file must be in pickDir
    if pickFile not in main.hotVar['pickFiles'].val:
        print(pickFile+' is not present in '+main.hotVar['pickDir'].val)
        return False
    return True

# Ensure that the pick set has the proper dimensions and data types [str,str,float] (although held as string)...
# ...also correct N.S.L format
def checkPickSet(main,pickSet):
    if len(pickSet.shape)!=2:
        print('The pickSet must be 2 dimensional')
        return False
    elif pickSet.shape[1]!=3:
        print('The pickSet must have 3 columns')
        return False
    try:
        pickSet[:,2].astype(float)
    except:
        print('The pickSet timestamp (third column) must contain numbers')
        return False
    # Check that the station string (N.S.L) has the correct format
    for entry in pickSet[:,0]:
        if not checkNSL(entry):
            return False
    return True

# Check that the station string (N.S.L) has the correct format
# Each entry (Network.Station.Location) in the list should be a string, and be max length of (2.5.2)
def checkNSL(nsl):
    splitEntry=nsl.split('.')
    if len(splitEntry)!=3:
        print('Station strings refer to Network.Station.Location')
        return False
    for string,maxLen,name in zip(splitEntry,[2,5,2],['Network','Station','Location']):
        if len(string)>maxLen:
            print('Station strings refer to Network.Station.Location; '+
                  name+' has max length of '+str(maxLen)+' characters')
            return False
    return True

# Ensure that the returned pen assignment (for traces) dictionary is holding the right data types
def checkTracePenAssign(main,penAssign):
    # Check to see that each returned value is a list
    for key,val in iteritems(penAssign):
        if type(val)!=list:
            print('For all tracePenAssign returned key:value pairs, the value should be a list')
            return False
        # Each entry (channel) in the list should be a string, and <= 3 characters long
        for entry in val:
            if type(entry) not in [str,np.string_,np.str_]:
                print('Returned tracePenAssign, lists should only contain strings, got type '+str(type(entry)))
                return False
            elif len(entry)>3:
                print('Returned tracePenAssign, list entries refer to channels, which are max 3 characters long')
                return False
    return True

# Ensure that the returned pen assignment (for stations) dictionary is holding the right data types
def checkStaColAssign(main,colAssign):
    # Check to see that each returned value is a list
    for key,val in iteritems(colAssign):
        if type(val)!=list:
            print('For all traceBgPenAssign/mapStaPenAssign returned key:value pairs, the value should be a list')
            return False
        # Check that the station string (N.S.L) has the correct format
        for entry in val:
            if type(entry) not in [str,np.string_,np.str_]:
                print('Returned traceBgPenAssign/mapStaPenAssign, lists should only contain strings, got type '+str(type(entry)))
                return False
            if not checkNSL(entry):
                return False
    return True
    
# Ensure that the current/previous event array is of the proper dimensions and datatype
def checkEveArr(main,eveArr):
    if len(eveArr.shape)!=2:
        print('The current/previous event array must be 2 dimensional')
        return False
    elif eveArr.shape[1]!=5:
        print('The current/previous event array must have 5 columns')
        return False
    try:
        eveArr.astype(float)
    except:
        print('The current/previous event array contains [ID,X,Y,Z,Timestamp(s)], which must all be numbers')
        return False
    return True

# Ensure that the mapPolygon array is of the proper dimensions and datatype
def checkMapPolygon(main,mapPolygon):
    if len(mapPolygon.shape)!=2:
        print('The mapPolygon array must be 2 dimensional')
        return False
    elif mapPolygon.shape[1]!=2:
        print('The mapPolygon array must have 2 columns')
        return False
    try:
        mapPolygon.astype(float)
    except:
        print('The mapPolygon array contains a list of vertices [[X1,Y1],[X2,Y2],...], which must all be numbers')
        return False
    return True

# Ensure that the archive directory exists  
def checkArchDir(main,archDir):
    if not os.path.isdir(archDir):
        print('Archive directory does not exist')
        return False
    return True
    
# Ensure that the station file exists and is in the proper format
def checkStaFile(main,staFile):
    # First see if the file exists
    if not os.path.isfile(staFile):
        print('Station file does not exist')
        return False
    # Ensure the to-read file is of proper format
    try:
        read_inventory(staFile,format='stationxml')
    except:
        print('Station file was not in xml format')
        return False
    return True

# Ensure that the image contains proper key words, and are in correct format
def checkImage(main,image):
    # Let user know if they gave useless keys
    acceptKeys=['data','t0','y0','tDelta','yDelta','label','cmapPos','cmapRGBA']
    givenKeys=image.keys()
    for key in givenKeys:
        if key not in acceptKeys:
            print('Image dict contained key '+key+' which is not used, accepted keys: '+str(acceptKeys))
    # Check to see that the forced key words are present
    keyFail=False
    for key in ['data','t0','tDelta']:
        if key not in givenKeys:
            print('Image key '+key+' was not contained in image dictionary')
            keyFail=True
    if keyFail:
        return False
    npArrType=type(np.array([0.0]))
    # Check to see that all arguments are of the proper type and dimensions
    for key,dtypes in [['data',[npArrType]],['t0',[float,np.float_,int,np.int_]],['y0',[float,np.float_,int,np.int_]],
                      ['tDelta',[float,np.float_,int,np.int_]],['yDelta',[float,np.float_,int,np.int_]],
                      ['label',[str,np.string_,np.str_,unicode]],['cmapPos',[npArrType]],['cmapRGBA',[npArrType]]]:
        if key in givenKeys:
            if type(image[key]) not in dtypes:
                print('Image key '+key+' has the expected '+str(dtypes[0])+' but was '+str(type(image[key])))
                keyFail=True
    if keyFail:
        return False
    # Check that the delta values are positive values
    for key in ['tDelta','yDelta']:
        if key in givenKeys:
            if image[key]<=0:
                keyFail=True
    if keyFail:
        print('Image delta values must be positive numbers')
        return False
    # Check the dimensions and contents of the numpy arrays
    if len(image['data'].shape)!=2 or 0 in image['data'].shape:
        print('Image data was not 2-dimensional, or had 0 length in at least one axis')
        keyFail=True
    if 'cmapPos' in givenKeys:
        if len(image['cmapPos'].shape)!=1:
            print('Image cmapPos was not 1-dimensional, or had length less than 2')
            keyFail=True
    if 'cmapRGBA' in givenKeys:
        if len(image['cmapRGBA'].shape)!=2:
            print('Image cmapRGBA was not 2-dimensional, or had length less than 2')
            keyFail=True
        elif image['cmapRGBA'].shape[1]!=4:
            print('Image cmapRGBA rows should contain length-4 arrays of the RGBA values')
            keyFail=True
        elif np.min(image['cmapRGBA'])<0 or np.max(image['cmapRGBA'])>255:
            print('Image cmapRGBA values should all be within the bounds [0,255]')
            keyFail=True
    if 'cmapRGBA' in givenKeys and 'cmapPos' in givenKeys:
        if image['cmapPos'].shape[0]!=image['cmapRGBA'].shape[0]:
            print('cmapPos and cmapRGBA must contain the same number of rows')
            keyFail=True
    if keyFail:
        return False
    return True


# Author: Andrew.M.G.Reynen
from future.utils import iteritems

import numpy as np
from obspy import UTCDateTime

# Convert a string into easy to read text, assumes no lists
def dict2Text(aDict):  
    # Get all values first as strings
    keys=np.array([key for key,val in iteritems(aDict)])
    vals=np.array([str(val) for key,val in iteritems(aDict)])
    # Sort them alphabetically (better than random atleast)
    argSort=np.argsort(keys)
    vals=vals[argSort]
    keys=keys[argSort]
    # Put them all into one larger string   
    string=''
    for i in range(len(vals)):
        string+=keys[i]+'='+vals[i]+',' 
    if len(string)>0: 
        string=string[:-1]
    return string
    
# Convert text into a dictionary,assumes no lists
def text2Dict(text):
    if '=' not in text:
        return {}
    aDict=dict(x.split('=') for x in text.split(','))
    # Check if any of the values represent type other than string
    for key,val in iteritems(aDict):
        # Do not convert numbers to bool
        if val=='True':
            aDict[key]=True
        elif val=='False':
            aDict[key]=False
        # Otherwise see if it is a number
        else:
            try:
                if '.' in val:
                    aDict[key]=float(val)
                else:
                    aDict[key]=int(val)
            # If none of the above, leave as a string
            except:
                aDict[key]=str(val)
    return aDict
    
# Return the UTCDateTime from the forced file naming convention
def getTimeFromFileName(fileName):
    timeStr=fileName.split('_')[1].replace('.'+fileName.split('.')[-1],'')
    return UTCDateTime().strptime(timeStr,'%Y%m%d.%H%M%S.%f')


# Get the string representing the given station via the trace object
def getStaStr(tr):
    return str('.'.join([tr.stats.network,tr.stats.station,tr.stats.location]))


# Author: Andrew.M.G.Reynen
from __future__ import print_function, division
from decimal import Decimal

import numpy as np
from obspy import UTCDateTime
from PyQt5 import QtWidgets, QtGui, QtCore
import pyqtgraph as pg
from CustomFunctions import getTimeFromFileName

# Custom axis labels for the archive widget
class TimeAxisItemArchive(pg.AxisItem):
    def __init__(self, *args, **kwargs):
        super(TimeAxisItemArchive, self).__init__(*args, **kwargs)
        # Some nice units with respect to time
        self.units=np.array([60,
                             120,300,600,1800,3600,7200,
                             3600*6,3600*24,3600*48,3600*24*5,3600*24*31,3600*24*180],dtype=float)
        self.unitIdxs=np.arange(len(self.units))
        # What the above units should be displayed as
        self.unitStrs=['%Hh%Mm',
                       '%Hh%Mm','%Hh%Mm','%Hh%Mm','%Hh%Mm','%Y-%m-%d %Hh','%Y-%m-%d %Hh',
                       '%Y-%m-%d %Hh','%Y-%m-%d','%Y-%m-%d','%Y-%m-%d','%Y-%m-%d','%Y-%m-%d']
        self.unitIdx=0

    # Customize the locations of the ticks
    def tickValues(self, minVal, maxVal, size):
        # How many ticks are wanted on the page
        numTicks=float(max([int(size/225),2]))
        # Given how big the page is, see which units to apply
        diff=maxVal-minVal
        self.unitIdx=int(np.interp((diff)/numTicks,self.units,self.unitIdxs))
        unit=self.units[self.unitIdx]
        # In the case where zoomed in way to close
        if unit>diff:
            return [(0.5*diff,[minVal+0.20*diff,maxVal-0.20*diff])]
        # In the case where zoomed out very far
        if diff/(unit*numTicks)>2:
            unit*=int(diff/(unit*numTicks))
        # Start at the nearest whole unit
        val=minVal-minVal%unit+unit
        ticks=[]
        while val<maxVal:
            ticks.append(val)
            val+=unit
        return [(unit,ticks)]

    # Customize what string are being shown
    def tickStrings(self, values, scale, spacing):
        return [UTCDateTime(value).strftime(self.unitStrs[self.unitIdx]) for value in values]

# Function called to display date time text on widget
def showHoverText(self,pixPoint):
    mousePoint=self.pltItem.vb.mapSceneToView(pixPoint)
    if np.isnan(mousePoint.x()):
        return
    self.hoverTimeItem.setText(str(UTCDateTime(Decimal(mousePoint.x()))))
    t1,t2=self.getPlotItem().vb.viewRange()[0]
    if t2<=1:
        return
    self.hoverTimeItem.setPos(t1+(t2-t1)*0.5,0)
    self.hoverTimeItem.show() 
    
# Hide the date time text when not hovering over widget
def hideHoverText(self,ev):
    super(pg.PlotWidget, self).leaveEvent(ev)
    self.hoverTimeItem.hide()
    
# Add a new pick file with the center mouse button
def addNewPickFile(self, ev):
    super(pg.PlotWidget, self).mouseDoubleClickEvent(ev)
    self.newEveTime=self.pltItem.vb.mapSceneToView(ev.pos()).x()
    self.addNewEventSignal.emit()
    
# Initiate common functions between the archiveEvent and archiveSpan widgets
def addArchiveWidgetElements(self):
    # Create text item
    self.hoverTimeItem=pg.TextItem(text='',anchor=(0.5,1))
    self.hoverTimeItem.setZValue(5)
    self.hoverTimeItem.hide()
    self.addItem(self.hoverTimeItem)
    # Add the show text function
    self.scene().sigMouseMoved.connect(lambda pixPoint:showHoverText(self,pixPoint))
    # Add the hide text function
    self.leaveEvent=lambda ev:hideHoverText(self,ev)
    
    # Add ability to add empty pick files if double clicked
    self.mouseDoubleClickEvent=lambda ev:addNewPickFile(self,ev)
    
# Widget for seeing what data times are available, and sub-selecting pick files
class ArchiveSpanWidget(pg.PlotWidget):
    addNewEventSignal=QtCore.pyqtSignal()
    
    def __init__(self, parent=None):
        super(ArchiveSpanWidget, self).__init__(parent,axisItems={'bottom': TimeAxisItemArchive(orientation='bottom')})
        self.pltItem=self.getPlotItem()
        self.pltItem.setMenuEnabled(enableMenu=False)
        # Turn off the auto ranging
        self.pltItem.vb.disableAutoRange()
        # Don't plot anything outside of the plot limits
        self.pltItem.setClipToView(True)
        # Only show the bottom axis
        self.pltItem.hideAxis('left')
        self.pltItem.hideButtons()
        # Give this widget a span select, for zooming
        self.span = pg.LinearRegionItem([500,1000])
        self.addItem(self.span)
        # Give a name to the data availability regions
        self.boxes=[]
        # Give some x-limits (do not go prior/after to 1970/2200)
        self.getPlotItem().setLimits(xMin=0,xMax=UTCDateTime('2200-01-01').timestamp)
        # No y-axis panning or zooming allowed
        self.pltItem.vb.setMouseEnabled(y=False)
        self.pltItem.setYRange(0,1)
        # Give some text upon hovering, and allow events to be added on double clicking
        addArchiveWidgetElements(self)
        
    # Return the span bounds
    def getSpanBounds(self):
        t1,t2=self.span.getRegion()
        return UTCDateTime(t1),UTCDateTime(t2)
    
    # Disable any scroll in/scroll out motions
    def wheelEvent(self,ev):
        return
        
    # Update the boxes representing the times
    def updateBoxes(self,ranges,penInt):
        # Remove the previous boxes
        for item in self.boxes:
            self.removeItem(item)
        self.boxes=[]
        # Skip, if no ranges to add
        if len(ranges)==0:
            return
        # Try to merge these ranges (less lines)...
        # ...sort first by range start values
        ranges=np.array(ranges)
        ranges=ranges[np.argsort(ranges[:,0])]
        merge=[ranges[0]]
        for rng in ranges:
            # In the case they do not overlap
            if rng[0]>merge[-1][1]:
                merge.append(rng)
            # If overlaping, and the new range extends further, push it
            elif rng[1]>merge[-1][1]:
                merge[-1][1]=rng[1]
        merge=np.array(merge)
        # Plot these efficiently (one disconnected line, rather than many lines)
        connect = np.ones((len(merge), 2), dtype=np.ubyte)
        connect[:,-1] = 0  #  disconnect segment between lines
        path = pg.arrayToQPath(merge.reshape(len(merge)*2), np.ones((len(merge)*2))*0.5, connect.reshape(len(merge)*2))
        item = pg.QtGui.QGraphicsPathItem(path)
        item.setPen(pg.mkPen(QtGui.QColor(penInt)))
        # Add in the new box
        self.boxes.append(item)
        self.addItem(item)      
        
# Graphview widget which holds all current pick files
class ArchiveEventWidget(pg.PlotWidget): 
    addNewEventSignal=QtCore.pyqtSignal()
    
    def __init__(self, parent=None):
        super(ArchiveEventWidget, self).__init__(parent)
        self.pltItem=self.getPlotItem()
        self.pltItem.setMenuEnabled(enableMenu=False)
        # Don't plot anything outside of the plot limits
        self.pltItem.setClipToView(True)
        # Do not show any axes
        self.pltItem.hideAxis('left')
        self.pltItem.hideAxis('bottom')
        self.pltItem.hideButtons()
        # Give some x-limits (do not go prior/after to 1970/2200)
        self.getPlotItem().setLimits(xMin=0,xMax=UTCDateTime('2200-01-01').timestamp,yMin=-0.3,yMax=1.3)
        # Disable all panning and zooming
        self.pltItem.vb.setMouseEnabled(x=False,y=False)
        self.curEveLine=None # For the highlighted event
        self.otherEveLine=None # For all other events
        # Give some text upon hovering, and allow events to be added on double clicking
        addArchiveWidgetElements(self)
    
    # Scale to where ever the archiveSpan has selected
    def updateXRange(self,linkWidget):
        self.setXRange(*linkWidget.getRegion(), padding=0)
        
    # Disable any scroll in/scroll out motions
    def wheelEvent(self,ev):
        return
    
    # Reset the event lines, given a new set of event times
    def updateEveLines(self,fileTimes,curFile):
        # Remove the not-current event lines
        if self.otherEveLine!=None:
            self.removeItem(self.otherEveLine)
            self.otherEveLine=None
        # Generate the many pick lines as one disconnected line
        if len(fileTimes)!=0:
            fileTimes=np.reshape(fileTimes,(len(fileTimes),1))
            times=np.hstack((fileTimes,fileTimes))
            connect = np.ones((len(times), 2), dtype=np.ubyte)
            connect[:,-1] = 0  #  disconnect segment between lines
            connect=connect.reshape(len(times)*2)
            path = pg.arrayToQPath(times.reshape(len(times)*2),connect,connect)
            item = pg.QtGui.QGraphicsPathItem(path)
            # Reference and plot the line
            self.otherEveLine=item
            self.addItem(item)
        # Generate and plot the current pick file (if present)
        self.updateEveLineSelect(curFile)
    
    # Update the current event line to proper position, and appropriate color
    def updateEveLineSelect(self,curFile):
        # Remove this line if no current file
        if self.curEveLine!=None:
            self.removeItem(self.curEveLine)
            self.curEveLine=None
        # Add the line if a pick file is selected
        if curFile!='':
            t=getTimeFromFileName(curFile)
            line=pg.InfiniteLine(pos=t,pen=pg.mkPen(QtGui.QColor(0)))
            # Reference and plot the line
            self.curEveLine=line
            self.addItem(line)
            
    # Update the color, width and depth of an event line
    def updateEvePens(self,evePen,eveType):
        col,width,dep=evePen
        col=QtGui.QColor(QtGui.QColor(col))
        if eveType=='cur':
            item=self.curEveLine
        else:
            item=self.otherEveLine
        if item!=None:
            item.setPen(pg.mkPen(col,width=width))
            item.setZValue(dep)

# List widget which holds all current pick files
class ArchiveListWidget(QtWidgets.QListWidget):       
    def __init__(self, parent=None):
        super(ArchiveListWidget, self).__init__(parent)
        
    # Return the list entries in the order which it appears
    def visualListOrder(self):
        aList=np.array([self.item(i).text() for i in range(self.count())])
        args=np.array([self.indexFromItem(self.item(i)).row() for i in range(self.count())])
        return [str(aFile) for aFile in aList[np.argsort(args)]]
        
    # Ensure that key presses are sent to the widget which the mouse is hovering over
    def enterEvent(self,ev):
        super(ArchiveListWidget, self).enterEvent(ev)
        self.setFocus()
        
    # Exit focus when the list widget is left
    def leaveEvent(self,ev):
        super(ArchiveListWidget, self).leaveEvent(ev)
        self.clearFocus()
        
# Custom axis labels for the time widget        
class TimeAxisItem(pg.AxisItem):
    def __init__(self, *args, **kwargs):
        super(TimeAxisItem, self).__init__(*args, **kwargs)
        # Some nice units with respect to time
        self.units=np.array([10**-6,10**-5,10**-4,10**-3,10**-2,10**-1,
                             1,2,5,10,30,60,
                             120,300,600,1800,3600,7200,
                             3600*6,3600*24,3600*48,3600*24*5,3600*24*31,3600*24*180],dtype=float)
        self.unitIdxs=np.arange(len(self.units))
        # What the above units should be displayed as
        self.unitStrs=['%S.%fs','%S.%fs','%S.%fs','%S.%fs','%S.%fs','%S.%fs',
                       '%S.%fs','%S.%fs','%Mm%Ss','%Mm%Ss','%Mm%Ss','%Mm%Ss',
                       '%Mm%Ss','%Hh%Mm','%Hh%Mm','%dd %Hh%Mm','%dd %Hh%Mm','%dd %Hh%Mm',
                       '%m-%d %Hh','%m-%d %Hh','%m-%d %Hh','%Y-%m-%d','%Y-%m-%d','%Y-%m-%d']
        # How many characters to trim off the end of the above string
        self.strTrim=[0,0,0,0,1,2,
                      3,3,0,0,0,0,
                      0,0,0,0,0,0,
                      0,0,0,0,0,0]
        self.unitIdx=0

    # Customize the locations of the ticks
    def tickValues(self, minVal, maxVal, size):
        # How many ticks are wanted on the page
        numTicks=float(max([int(size/160),2]))
        # Given how big the page is, see which units to apply
        diff=maxVal-minVal
        self.unitIdx=int(np.interp((diff)/numTicks,self.units,self.unitIdxs))
        unit=self.units[self.unitIdx]
        # In the case where zoomed in way to close
        if unit>diff:
            return [(0.5*diff,[minVal+0.20*diff,maxVal-0.20*diff])]
        # In the case where zoomed out very far
        if diff/(unit*numTicks)>2:
            unit*=int(diff/(unit*numTicks))
        # Start at the nearest whole unit
        val=minVal-minVal%unit+unit
        ticks=[]
        while val<maxVal:
            ticks.append(val)
            val+=unit
        return [(unit,ticks)]

    # Customize what string are being shown
    def tickStrings(self, values, scale, spacing):
        x=self.strTrim[self.unitIdx]
        if x==0:
            return [UTCDateTime(value).strftime(self.unitStrs[self.unitIdx]) for value in values]
        else:
            return [UTCDateTime(value).strftime(self.unitStrs[self.unitIdx])[:-(x+1)]+'s' for value in values]

# Widget to nicely show the time axis, which is in sync with the trace data
class TimeWidget(pg.PlotWidget):
    def __init__(self, parent=None):
        super(TimeWidget, self).__init__(parent,name='timeAxis',axisItems={'bottom': TimeAxisItem(orientation='bottom'),
                                                                           'left':pg.AxisItem(orientation='left',showValues=False)})
        self.getPlotItem().getAxis('left').setWidth(70)
        self.getPlotItem().setMenuEnabled(enableMenu=False)
        self.getPlotItem().hideButtons()
        # Give some x-limits (do not go prior/after to 1970/2200)
        self.getPlotItem().setLimits(xMin=0,xMax=UTCDateTime('2200-01-01').timestamp)
        self.getPlotItem().vb.setMouseEnabled(y=False,x=False)
        # Turn off all interaction
        self.setEnabled(False)
        # Give a blank title (gets rid of the y-axis)
        self.getPlotItem().setLabels(title='')
    
    # Null any built in hover actions
    def enterEvent(self,event):
        return
        
    def leaveEvent(self,event):
        return
    
    # Return the xmin,xmax (times) of the plot
    def getTimeRange(self):
        return self.getPlotItem().vb.viewRange()[0]

# Infinite line, but now with reference to the pick type
class PickLine(pg.InfiniteLine):
    def __init__(self, aTime,aType,pen,parent=None):
        super(PickLine, self).__init__(parent)
        col,width,depth=pen
        self.pickType=aType
        self.setValue(aTime)
        self.setZValue(depth)
        self.setPen(col,width=width)
        self.setVisible(width!=0)

# Plot curve item, but now with reference to the channel
class TraceCurve(pg.PlotCurveItem):
    def __init__(self,x,y,cha,pen,dep,parent=None):
        super(TraceCurve,self).__init__(parent)
        self.setData(x=x,y=y,pen=pen)
        self.cha=cha
        self.setZValue(dep)
        # If width of zero, do not show the curve
        if pen.widthF()==0:
            self.setVisible(False)

# Widget which will hold the trace data, and respond to picking keybinds     
class TraceWidget(pg.PlotWidget):
    doubleClickSignal=QtCore.pyqtSignal()  
    
    def __init__(self, parent=None,sta='',hoverPos=None):
        super(TraceWidget, self).__init__(parent)
        self.pltItem=self.getPlotItem()
        self.pltItem.setMenuEnabled(enableMenu=False)
        # Speed up the panning and zooming
        self.pltItem.setClipToView(True)
        self.pltItem.setDownsampling(True, True, 'peak')
        # Turn off the auto ranging
        self.pltItem.vb.disableAutoRange()
        # Only show the left axis
        self.pltItem.hideAxis('bottom')
        self.pltItem.hideButtons()
        self.pltItem.getAxis('left').setWidth(70)
        # Give some x-limits (do not go prior/after to 1970/2200)
        self.getPlotItem().setLimits(xMin=0,xMax=UTCDateTime('2200-01-01').timestamp)
        # Assign this widget a station
        self.sta=sta
        # Allow the widget to hold memory of pick lines, and traces
        self.pickLines=[]
        self.traceCurves=[]
        # Set the hover position, upon hovering
        self.scene().sigMouseMoved.connect(self.onHover)
        self.hoverPos=hoverPos
        # Allow this widget to be fairly small
        sizePolicy = QtGui.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.MinimumExpanding)
        self.setSizePolicy(sizePolicy)
        self.setMinimumSize(QtCore.QSize(0, 15))
        
    # Update the mouse position
    def onHover(self,pixPoint):
        mousePoint=self.pltItem.vb.mapSceneToView(pixPoint)
        self.hoverPos=Decimal(mousePoint.x()),Decimal(mousePoint.y())
    
    # Add a trace to the widget
    def addTrace(self,x,y,cha,pen,dep):
        curve=TraceCurve(x,y,cha,pen,dep)
        self.addItem(curve)
        self.traceCurves.append(curve)
    
    # Add a single pick line to this station
    def addPick(self,aTime,aType,pen):
        aLine=PickLine(aTime,aType,pen)
        # Add the line to the plot, and its own pick lines
        self.pltItem.addItem(aLine)
        self.pickLines.append(aLine)
        
    # Remove a specific number of picks from the pick lines (with a given pick type)
    def removePicks(self,aType,delTimes):
        # Count forwards to select picks for deletion
        delIdxs=[]
        for i,aLine in enumerate(self.pickLines):
            # If picks overlap, do not delete more than necessary
            if len(delIdxs)==len(delTimes):
                break
            if aLine.pickType==aType and aLine.value() in delTimes:
                delIdxs.append(i)
        # Check to make sure all were deleted
        if len(delIdxs)!=len(delTimes):
            print('Missed deleting some plotted pick lines')
        # Loop backwards and pop these picks off the widget
        for idx in delIdxs[::-1]:
            aLine=self.pickLines.pop(idx)
            self.pltItem.removeItem(aLine)
    
    # Emit signal for picking
    def mouseDoubleClickEvent(self, ev):
        super(TraceWidget, self).mouseDoubleClickEvent(ev)
        # If this widget is assigned a station, return signal
        if self.sta!=None:
            self.doubleClickSignal.emit()
    
    # Ensure that key presses are sent to the widget which the mouse is hovering over
    def enterEvent(self,ev):
        super(TraceWidget, self).enterEvent(ev)
        self.setFocus()
        
    # Exit focus when the trace is left
    def leaveEvent(self,ev):
        super(TraceWidget, self).leaveEvent(ev)
        self.clearFocus()
        self.hoverPos=None
        
    # Return the ymin,ymax of the plot
    def getRangeY(self):
        return self.getPlotItem().vb.viewRange()[1]
    
# Widget to hold raster information for plotting
class ImageWidget(pg.PlotWidget): 
    def __init__(self, parent=None):
        super(ImageWidget, self).__init__(parent)        
        self.pltItem=self.getPlotItem()
        self.pltItem.setMenuEnabled(enableMenu=False)
        # Speed up the panning and zooming
        self.pltItem.setClipToView(True)
        # Turn off the auto ranging
        self.pltItem.vb.disableAutoRange()
        # Only show the left axis
        self.pltItem.hideAxis('bottom')
        self.pltItem.hideButtons()
        self.pltItem.getAxis('left').setWidth(70)
        self.pltItem.getAxis('left').setZValue(1)
        # Give some x-limits (do not go prior/after to 1970/2200)
        self.getPlotItem().setLimits(xMin=0,xMax=UTCDateTime('2200-01-01').timestamp)
        # Create a blank image item
        self.imageItem=pg.ImageItem()
        self.addItem(self.imageItem)
        self.prevPosX,self.prevScaleX=0,1
        self.prevPosY,self.prevScaleY=0,1
        # Give the image a border
        self.imageBorder=pg.QtGui.QGraphicsRectItem(-1,-1,0.1,0.1)
        self.addItem(self.imageBorder)
        
    # Update the image on this widget
    def loadImage(self,imgDict):
        # Load in the defaults if some optional keys are not present
        imgDict=self.loadDefaultKeys(imgDict)
        # Set the image data
        self.imageItem.setImage(imgDict['data'].T)
        # Set position and scale of image...
        diffX=imgDict['t0']-self.prevPosX
        diffY=imgDict['y0']-self.prevPosY
        self.imageItem.scale(imgDict['tDelta']/self.prevScaleX,imgDict['yDelta']/self.prevScaleY)
        self.imageItem.translate(diffX/imgDict['tDelta'], diffY/imgDict['yDelta'])
        self.prevPosX+=diffX
        self.prevPosY+=diffY
        self.prevScaleX,self.prevScaleY=imgDict['tDelta'],imgDict['yDelta']
        # Set the coloring
        lut=self.getLUT(imgDict['data'],imgDict['cmapPos'],imgDict['cmapRGBA']) 
        self.imageItem.setLookupTable(lut)
        # Update image boundary
        xSize,ySize=np.array([imgDict['tDelta'],imgDict['yDelta']])*imgDict['data'].T.shape
        self.imageBorder.setRect(imgDict['t0'],imgDict['y0'],xSize,ySize)
        # Set default Y-limit to zoom to the image
        self.setYRange(imgDict['y0'],ySize,padding=0.0)
        # Apply the label
        self.pltItem.setLabel(axis='left',text=imgDict['label'])
    
    # Give default values to keys which are not present in the image dictionary
    def loadDefaultKeys(self,imgDict):
        for key,val in [['y0',0],['yDelta',1],['label',''],
                        ['cmapPos',np.array([np.min(imgDict['data']),np.max(imgDict['data'])])],
                        ['cmapRGBA',np.array([[0,0,0,255],[255,255,255,255]])]
                        ]:
            if key not in imgDict.keys():
                # In the case where the colors are given, but the positions are not, give evenly spaced positions
                if key=='cmapPos' and 'cmapRGBA' in imgDict.keys():
                    imgDict[key]=np.linspace(np.min(imgDict['data']),np.max(imgDict['data']),len(imgDict['cmapRGBA']))
                else:
                    imgDict[key]=val
        # In the case where the positions are given, but not the colors - just use first and last position
        if len(imgDict['cmapPos'])!=len(imgDict['cmapRGBA']):
            imgDict['cmapPos']=np.array([imgDict['cmapPos'][0],imgDict['cmapPos'][-1]])
        return imgDict
    
    # Get the look-up-table (color map) 
    def getLUT(self,data,pos,col):
        aMap = pg.ColorMap(np.array(pos), np.array(col,dtype=np.ubyte))
        lut = aMap.getLookupTable(np.min(data),np.max(data),256)
        return lut


# Author: Andrew.M.G.Reynen
from __future__ import print_function
import os

from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt

from ChangeSource import Ui_CsDialog

# Saved sources class, for reading in old (or adding new) archive/pick/station information
class SaveSource(object):
    def __init__(self,tag=None,archDir=None,
                 pickDir=None,staFile=None,
                 sourceDict={}):
        self.tag=tag
        self.archDir=archDir
        self.pickDir=pickDir
        self.staFile=staFile
        self.sourceDict=sourceDict
    
    # A check to see if all files can be read
    def pathExist(self):
        allPathsExist=True
        for txt,path in [['archive directory',self.archDir],
                         ['pick directory',self.pickDir],
                         ['station file',self.staFile]]:
            if not os.path.exists(path):
                # Allow for no station file to be read
                if txt=='station file' and path.replace(' ','')=='':
                    continue
                print('The '+txt+' does not exist at: '+path)
                allPathsExist=False
        return allPathsExist


# Default saved source
def defaultSource():
    exampleDir=os.path.dirname(os.path.realpath(__file__))+'/ExampleSource/'
    source=SaveSource(tag='ExampleSource',
                      archDir=exampleDir+'Archive',
                      pickDir=exampleDir+'Picks',
                      staFile=exampleDir+'NX_stations.xml',
                      sourceDict={'Velocity':{'Vp':6.19,'Vs':3.57,
                                              'Dp':0.74,'Ds':1.13,
                                              'tResThresh':1.0}})
    return {source.tag:source}

# Change source dialog
class CsDialog(QtWidgets.QDialog, Ui_CsDialog):
    def __init__(self,hotVar,savedSources,parent=None):
        QtWidgets.QDialog.__init__(self,parent)
        self.setupUi(self)
        self.hotVar=hotVar
        self.saveSource=savedSources
        # Give the dialog some functionaly
        self.setFunctionality()
        # Load in the text of the current source
        self.fillDialog()
        # Show the list of previous saved sources
        self.csSaveSourceList.addItems(sorted([key for key in self.saveSource.keys()]))
        
    # Set up some functionality to the configuration dialog
    def setFunctionality(self):
        self.csSaveSourceList.itemDoubleClicked.connect(self.loadSaveSource)
        self.csSaveSourceList.keyPressedSignal.connect(self.delSaveSource)
        self.csSaveSourceButton.clicked.connect(self.addSavedSource)
        # Allow double click on paths to open up dialogs to extract path names
        self.csArchiveLineEdit.doubleClicked.connect(lambda: self.getPathName('arch'))
        self.csPickLineEdit.doubleClicked.connect(lambda: self.getPathName('pick'))
        self.csStationLineEdit.doubleClicked.connect(self.getFileName)
        
    # Fill the dialog with info relating to current source
    def fillDialog(self):
        # Fill line edits with current source
        self.csTagLineEdit.setText(self.hotVar['sourceTag'].val)
        self.csArchiveLineEdit.setText(self.hotVar['archDir'].val)
        self.csPickLineEdit.setText(self.hotVar['pickDir'].val)
        self.csStationLineEdit.setText(self.hotVar['staFile'].val)
        self.sourceDictWidget.showSourceDict(self.hotVar['sourceDict'].val)
        
    # Put the saved source into the saved source list and dictionary
    def loadSaveSource(self):
        if self.csSaveSourceList.currentItem() is not None:
            if self.csSaveSourceList.currentItem().isSelected():
                source=self.saveSource[self.csSaveSourceList.currentItem().text()]
                self.csTagLineEdit.setText(source.tag)
                self.csArchiveLineEdit.setText(source.archDir)
                self.csPickLineEdit.setText(source.pickDir)
                self.csStationLineEdit.setText(source.staFile)
                # If the source had no source dict, use the default
                if hasattr(source,'sourceDict'):
                    self.sourceDictWidget.showSourceDict(source.sourceDict)
                else:
                    self.sourceDictWidget.showSourceDict(SaveSource().sourceDict)
    
    # Delete the selected saved source
    def delSaveSource(self):
        if self.csSaveSourceList.key != Qt.Key_Delete or self.csSaveSourceList.currentItem()==None:
            return
        tag=self.csSaveSourceList.currentItem().text()
        # Remove from the saved sources dictionary...
        self.saveSource.pop(tag)
        # ...and the gui list
        self.csSaveSourceList.takeItem(self.csSaveSourceList.currentRow())
    
    # Save the edits to the currently selected source
    def addSavedSource(self):
        source=self.curSource()
        # Ensure that atleast the tag, archive directory, and pick directory are filled out
        for text in [source.tag,source.archDir,source.pickDir]:
            if text.replace(' ','')=='':
                print('Fill in the source information to save')
                return
        # Add this to the saved sources list
        if source.tag not in [key for key in self.saveSource.keys()]:
            self.csSaveSourceList.addItem(source.tag)
        self.saveSource[source.tag]=source
        
    # Function to open a dialog and get the path name
    def getPathName(self,thisLineEdit):
        # Check which line was clicked
        if thisLineEdit=='arch':
            line=self.csArchiveLineEdit
        else:
            line=self.csPickLineEdit
        # If the value is already set, start from there
        if os.path.isdir(line.text()):
            startFolder=line.text()
        else:
            startFolder=os.path.dirname(os.path.realpath(__file__))
        name=str(QtWidgets.QFileDialog.getExistingDirectory(self, "Select Folder",startFolder))
        name=name.replace('\\','/')
        # If a folder was selected, update the line edit
        if name!='' and os.path.isdir(name):
            line.setText(name)
            
    # Function to open a dialog and get the file name
    def getFileName(self):
        # Start the dialog from the previously selected file (if it exists)
        if os.path.isfile(self.csStationLineEdit.text()):
            startFolder=os.path.dirname(os.path.realpath(self.csStationLineEdit.text()))
        else:
            startFolder=os.path.dirname(os.path.realpath(__file__))
        name=str(QtWidgets.QFileDialog.getOpenFileName(self, "Select File",startFolder)[0])
        # If a folder was selected, update the line edit
        if name!='' and os.path.isfile(name):
            self.csStationLineEdit.setText(name)
        
    # Create a source object from the line edit
    def curSource(self):
        source=SaveSource(tag=self.csTagLineEdit.text(),
                          archDir=self.csArchiveLineEdit.text(),
                          pickDir=self.csPickLineEdit.text(),
                          staFile=self.csStationLineEdit.text(),
                          sourceDict=self.sourceDictWidget.getSourceDict())
        return source
        
    # Upon close, return the source currently in the line edits
    def returnSource(self):
        return self.curSource()